unit hxInstallUtilities;
{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

  // --------------------------------------------------------------------------
//written by handX software
//Copyright (C) 2001 handX software

//    This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU Lesser General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.

//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

//    You should have received a copy of the GNU Lesser General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// --------------------------------------------------------------------------
//
// If you have questions, improvements or other feedback, please contact
// support@handx.net.  You can also find discussion of this product hosted on
// the TurboSync yahoo group.  Visit http://groups.yahoo.com/group/TurboSync
//
// --------------------------------------------------------------------------


interface

Uses
  SysUtils, classes, shellLinks, ShlObj, Math, XMLIntf, hXxmlUtils, forms, windows;

type
  EConduitRegistryError = class(Exception);
  EConduitRegisteredError = class(Exception);
  EBadSettingsFileError = class(Exception);
  TIOResult = (ioRebootRequired, ioSuccess, ioFailed);
  TConduitType =  (ctStandard, ctJava, ctCom);


function addBackslashes(input: String): String;
function CopyInUseSysFile(const Source : String; const Dest: String): boolean;
function extensionIsPalm(extension: String) : boolean;
procedure FileCopy(sourceFile, destFile: String);
function findRegEntry(basePath, dllName: String; conduitType: TConduitType = ctStandard): String;
function stringToConduitType(input: string): TConduitType;
function conduitTypeToString(input: TConduitType): string;
function getExistingParentDirectory(dir: string): string;
function getFirstNonExistingDirectory(dir: string): string;
procedure GetFileVersion(FileName: string; var Major1, Major2,
         Minor1, Minor2: Integer);
procedure decreaseUsageCount(fileName: string; var rebootRequired);
procedure increaseUsageCount(fileName: string);
function GetFileVersionString(FileName: string): String;
function getFirstExistingRoot(directoryName: string): String;
function getFileSize(fileName: String): extended;
function getPrintableFileSize(fileName: String): string;
function fileSizeToPrintable(fileSize: integer): String;
function fileVersionIsHigher(compareFileName, againstFileName: String): Boolean;
function InstallFile(Source, Dest: String): TIOResult;
function IsFileInUse(FileName: String): Boolean;
function isNT: boolean;
function KillFile(fileName: String): TIOResult;
procedure KillDirectory(dirName: String);
function isDirectory(dirName: String): boolean;
function ProcessParam(Param: String): String;
function ShellExec(op, fn, par, dir: PChar; show: Word; wait: Boolean): LongInt;
procedure createLink(MenuPosition, fileName, Description: string);
function getStartMenuPath: string;
function getStartProgramsMenuPath: string;
function getWindowsPath: string;
function getSystemPath: string;
function getProgramFilesPath: string;
function getTempDir: string;
//procedure FindFiles(APath: String; FileMask: String; var FileList: TStringList);
function StringToHKey(input: String): HKey;
function NTSetPrivilege(sPrivilege: string; bEnabled: Boolean): Boolean;

implementation
uses
  Registry,
  IniFiles,
  shellapi;

function NTSetPrivilege(sPrivilege: string; bEnabled: Boolean): Boolean;
var
  hToken: THandle;
  TokenPriv: TOKEN_PRIVILEGES;
  PrevTokenPriv: TOKEN_PRIVILEGES;
  ReturnLength: Cardinal;
begin
  Result := True;
  // Only for Windows NT/2000/XP and later.
  if not (Win32Platform = VER_PLATFORM_WIN32_NT) then Exit;
//  Result := False;

  // obtain the processes token
  if OpenProcessToken(GetCurrentProcess(),
    TOKEN_ADJUST_PRIVILEGES or TOKEN_QUERY, hToken) then
  begin
    try
      // Get the locally unique identifier (LUID) .
      if LookupPrivilegeValue(nil, PChar(sPrivilege),
        TokenPriv.Privileges[0].Luid) then
      begin
        TokenPriv.PrivilegeCount := 1; // one privilege to set

        case bEnabled of
          True: TokenPriv.Privileges[0].Attributes  := SE_PRIVILEGE_ENABLED;
          False: TokenPriv.Privileges[0].Attributes := 0;
        end;

        ReturnLength := 0; // replaces a var parameter
        PrevTokenPriv := TokenPriv;

        // enable or disable the privilege

        AdjustTokenPrivileges(hToken, False, TokenPriv, SizeOf(PrevTokenPriv),
          PrevTokenPriv, ReturnLength);
      end;
    finally
      CloseHandle(hToken);
    end;
  end;
  // test the return value of AdjustTokenPrivileges.
  Result := GetLastError = ERROR_SUCCESS;
  if not Result then
    raise Exception.Create(SysErrorMessage(GetLastError));
end;

function addBackslashes(input: String): String;
begin
  if input = '' then
    begin
    result := '';
    exit
    end;
  if input[1] <> '\' then
    begin
    result := IncludeTrailingPathDelimiter('\' + input);
    end // if input[length(input)] <> '\'...
  else
    begin
    result := IncludeTrailingPathDelimiter(input);
    end;
end;

procedure decreaseUsageCount(fileName: string; var rebootRequired);
var
  reg: TRegistry;
  useCount: integer;
begin
  reg := tregistry.create(key_read or key_write);
  reg.RootKey := HKEY_LOCAL_MACHINE;
  reg.OpenKey('\Software\Microsoft\Windows\CurrentVersion\SharedDLLs', true);
  if reg.ValueExists(fileName) then
    begin
    useCount := reg.ReadInteger(fileName);
    if useCount <= 1 then
      begin
      if fileExists(fileName) then
        killFile(fileName);
      reg.DeleteValue(fileName);
      end //if useCount <= 1 then
    else
      begin
      dec(useCount);
      reg.WriteInteger(fileName, useCount);
      end;////if useCount <= 1 then.. else
    end //if reg.ValueExists(fileName) then
  else
    begin
    killFile(fileName);
    end; //if reg.ValueExists(fileName) then
  reg.Free;
end;

procedure increaseUsageCount(fileName: string);
var
  reg: TRegistry;
  useCount: integer;
begin
  reg := tregistry.create(key_read or key_write);
  reg.RootKey := HKEY_LOCAL_MACHINE;
  useCount := 0;
  reg.OpenKey('\Software\Microsoft\Windows\CurrentVersion\SharedDLLs', true);
  if reg.ValueExists(fileName) then
    begin
    useCount := reg.ReadInteger(fileName);
    inc(useCount);
    end;
  reg.WriteInteger(fileName, useCount);

  reg.Free;
end;

function getFirstNonExistingDirectory(dir: string): string;
begin
  Dir := ExcludeTrailingPathDelimiter(Dir);
{$IFDEF MSWINDOWS}
  if (Length(Dir) < 3) or DirectoryExists(Dir)
    or (ExtractFilePath(Dir) = Dir) then
      begin
      result := dir;
      exit;
      end;// avoid 'xyz:\' problem.
{$ENDIF}
{$IFDEF LINUX}
  if (Dir = '') or DirectoryExists(Dir) then
    begin
    result := dir;
    Exit;
    end;
{$ENDIF}
  if directoryExists(extractFilePath(dir)) then result := dir
  else result := getFirstNonExistingDirectory(extractFilePath(dir));
end;

function getExistingParentDirectory(dir: string): string;
begin
  Dir := ExcludeTrailingPathDelimiter(Dir);
{$IFDEF MSWINDOWS}
  if (Length(Dir) < 3) or DirectoryExists(Dir)
    or (ExtractFilePath(Dir) = Dir) then
      begin
      result := dir;
      exit;
      end;// avoid 'xyz:\' problem.
{$ENDIF}
{$IFDEF LINUX}
  if (Dir = '') or DirectoryExists(Dir) then
    begin
    result := dir;
    Exit;
    end;
{$ENDIF}
  Result := getExistingParentDirectory(extractFilePath(dir));
end;

function getFileSize(fileName: String): extended;
begin
  result := 0;
  if not fileExists(fileName) then exit;
  with TFileStream.Create(filename, fmOpenRead or fmShareDenyNone) do
    try
      result := Size;
    finally;
      free;
    end;
end;

function fileSizeToPrintable(fileSize: integer): String;
var
  fSize: extended;
begin
  fSize := fileSize;
  if fSize < 1024 then
    result := currToStr(fSize) + ' bytes'
  else if fSize > 1073741824 then
    begin
    fSize := fSize / 1073741824;
    fSize := roundTo(fSize, -1);
    result := currToStr(fSize) + ' GB';
    end
  else if fSize > 1048576 then
    begin
    fSize := fSize / 1048576;
    fSize := roundTo(fSize, -1);
    result := currToStr(fSize) + ' MB';
    end
  else
    begin
    fSize := fSize / 1024;
    fSize := roundTo(fSize, -1);
    result := currToStr(fSize) + ' KB';
    end;
end;

function getPrintableFileSize(fileName: String): string;
var
  fSize: Extended;
begin
  result := '';
  if not fileExists(fileName) then exit;
  fSize := getFileSize(fileName);
  result := fileSizeToPrintable(trunc(fSize));
end;

function StringToHKey(input: String): HKey;
begin
  if uppercase(input) = 'HKEY_LOCAL_MACHINE' then
    result := HKEY_LOCAL_MACHINE
  else if uppercase(input) = 'HKEY_CLASSES_ROOT' then
    result := HKEY_CLASSES_ROOT
  else if uppercase(input) = 'HKEY_USERS' then
    result := HKEY_USERS
  else if uppercase(input) = 'HKEY_CURRENT_CONFIG' then
    result := HKEY_CURRENT_CONFIG
  else
    result := HKEY_CURRENT_USER;

end;

function replaceEnvironmentVariable(input: String): string;
var
  beginPosition: integer;
  endPosition: integer;
  rplString: string;
begin
  beginPosition := pos('%', input);
  rplString := copy(input, beginPosition +1, length(input));
  endPosition := pos('%', rplString);
  rplString := copy(rplString, 1, endPosition -1);
  rplString := GetEnvironmentVariable(rplString);
  delete(input, beginPosition, endPosition + 1);
  insert(rplString, input, beginPosition);
  result := input;
end;

function getStartProgramsMenuPath: string;
var
  PIDL: PItemIDList;
  Path: array[0..MAX_PATH] of char;
begin
  SHGetSpecialFolderLocation(Application.Handle, CSIDL_COMMON_PROGRAMS, PIDL);
  if not SHGetPathFromIDList(PIDL, Path) then
    begin
    SHGetSpecialFolderLocation(Application.Handle, CSIDL_PROGRAMS, PIDL);
    SHGetPathFromIDList(PIDL, Path);
    end;
  result := Trim(String(path));
end;

function getStartMenuPath: string;
var
  PIDL: PItemIDList;
  Path: array[0..MAX_PATH] of char;
begin
  SHGetSpecialFolderLocation(Application.Handle, CSIDL_COMMON_STARTMENU, PIDL);
  if not SHGetPathFromIDList(PIDL, Path) then
    begin
    SHGetSpecialFolderLocation(Application.Handle, CSIDL_STARTMENU, PIDL);
    SHGetPathFromIDList(PIDL, Path);
    end;
  result := Trim(String(path));
end;

function getWindowsPath: string;
var
  Path : Array[0..MAX_PATH] of Char;
begin
  GetWindowsDirectory(Path, MAX_PATH);
  result := StrPas(Path);
end;

function getSystemPath: string;
var
  Path : Array[0..MAX_PATH] of Char;
begin
  GetSystemDirectory(Path, MAX_PATH);
  result := StrPas(Path);
end;

function getTempDir: string;
var
  Path : Array[0..MAX_PATH] of Char;
begin
  GetTempPath(MAX_PATH, Path);
  result := StrPas(Path);
end;


function getProgramFilesPath: string;
var
  reg: TRegistry;
begin
  reg := TRegistry.Create(key_read);
  reg.RootKey := HKEY_LOCAL_MACHINE;
  reg.OpenKey('Software\Microsoft\Windows\CurrentVersion', false);
  result := reg.ReadString('ProgramFilesDir');
  reg.Free;
end;

procedure createLink(MenuPosition, fileName, Description: string);
var
  lnk: TShellLink;
  dir: String;
begin
  dir := IncludeTrailingPathDelimiter(IncludeTrailingPathDelimiter(Trim(getStartProgramsMenuPath)) + MenuPosition);
  ForceDirectories(dir);
  lnk := TShellLink.Create(dir + description + '.lnk');
  lnk.setDescription(PChar(Description));
  lnk.setPath(fileName);
  lnk.Save;
  lnk.free;
end;

function CopyInUseSysFile(const Source : String; const Dest: String): boolean;
{Win95 accepts long File Name, however they can not be copied in MS-DOS
mode. So, we need to manipulate with short File Names.
GetShortPathName returns the length, in characters, of ShortPath
If there is no need to shorten the LongPath, ShortPath = LongPath}
var
   ShortSource, ShortDest : String;
   TempFile: String;
   WinInitINI : TIniFile;
   Attr: Integer;
begin
  try
    SetLength(ShortSource, MAX_PATH);
    SetLength(ShortDest, MAX_PATH);
    SetLength(TempFile, MAX_PATH);

    GetShortPathName(PChar(Source), PChar(ShortSource), MAX_PATH);
    GetShortPathName(PChar(Dest), PChar(ShortDest), MAX_PATH);
    GetTempPath(MAX_PATH, PChar(TempFile));
    TempFile:=strcat(PChar(TempFile), PChar(ExtractFileName(Source)));

    FileCopy(ShortSource, TempFile);
    Attr := FileGetAttr(TempFile);
    Attr := Attr and not(faReadOnly);
    FileSetAttr(TempFile, Attr);

    WinInitINI := TiniFile.Create('Wininit.ini');
    WinInitINI.WriteString('rename', ShortDest,
                         TempFile);
    WinInitINI.Free;
    result := true;
  except
    result := false;
  end;
end;

function extensionIsPalm(extension: String) : boolean;
begin
  result := (lowercase(extension) = '.pdb') or
        (lowercase(extension) = '.prc') or
        (lowercase(extension) = '.pqa') or
        (lowercase(extension) = '.pnc') or
        (lowercase(extension) = '.scp');

end;

procedure FileCopy(sourceFile, destFile: String);
var
  optStruct    : TShFileOpStruct;
  dest, source : PChar;
begin
  if lowerCase(sourceFile) = lowerCase(destFile) then exit;
  GetMem(source, Length(sourceFile)+2);
  try
    GetMem(dest, Length(destFile)+2);
    try
      FillChar(source^, Length(sourceFile)+2, 0);
      FillChar(dest^, Length(destFile)+2, 0);

      StrCopy(source, PChar(sourceFile));
      StrCopy(dest, PChar(destFile));

      with optStruct do
        begin
        Wnd := Application.Handle;
        wFunc := FO_COPY;
        pFrom := source;
        pTo := dest;
        fFlags := FOF_NOCONFIRMATION or FOF_NOCONFIRMMKDIR or FOF_SILENT;
        fAnyOperationsAborted := false;
        hNameMappings := nil;
        lpszProgressTitle := 'Updating Files';
        if (SHFileOperation(optStruct) <> 0) and (fAnyOperationsAborted = false) then
          begin
          RaiseLastOSError;
          end; // if SHFileOperation(optStruct) <>> 0...
        end;    // with
      finally
        FreeMem(dest, Length(destFile) + 2);
      end;
    finally
      FreeMem(source, Length(sourceFile) + 2);
    end;
end;

function fileVersionIsHigher(compareFileName, againstFileName: String): Boolean;
var
  c1,c2,c3,c4:integer;
  a1,a2,a3,a4:integer;
begin
  if not fileExists(againstFileName) then
    begin
    result := true;
    exit;
    end;
  getfileVersion(compareFileName,c1,c2,c3,c4);
  getfileVersion(againstFileName,a1,a2,a3,a4);
  result := (c1 > a1) or ((c1 = a1) and (c2 > a2)) or ((c1=a1) and (c2 = a2) and (c3>a3)) or
            ((c1=a1) and (c2 = a2) and (c2 = a2) and (c4>a4))
end;

(*procedure FindFiles(APath: String; FileMask: String; var FileList: TStringList);
{ This is a procedure which is called recursively so that it finds the
file with a specified mask through the current directory and its
sub-directories. }
var
  FSearchRec, DSearchRec: TSearchRec;
  FindResult: integer;
  subList : TStringList;
begin
  Application.ProcessMessages;
  APath := AddBackSlash(APath); // Obtain a valid directory name
  { Find the first occurance of the specified file name }
  FindResult := FindFirst(APath+FileMask,faAnyFile+faHidden+
                          faSysFile+faReadOnly,FSearchRec);
  try
    { Continue to search for the files according to the specified
      mask. If found, add the files and their paths to the stringlist.}
    while (FindResult = 0) do
      begin
      if (FileMask = '*.*') or (pos(copy(FileMask,2,length(FileMask) -2), FSearchRec.Name)=length(FSearchRec.Name) - length(copy(FileMask,2,length(FileMask) -2))) then
        begin
        FileList.Add(LowerCase(APath+FSearchRec.Name));
        end;
      FindResult := FindNext(FSearchRec);
      end; //WHILE
{ Now search the sub-directories of this current directory. Do this
by using FindFirst to loop through each subdirectory, then call
FindFiles (this function) again. This recursive process will
continue until all sub-directories have been searched. }
  FindResult := FindFirst(APath+'*.*', faDirectory, DSearchRec);
  while (FindResult = 0) do
    begin
    if ((DSearchRec.Attr and faDirectory)=faDirectory) and (pos('.', DSearchRec.Name)=0) then
      begin
      subList := TStringList.create;
      FindFiles(APath+DSearchRec.Name, fileMask, subList); // Recursion here
      FileList.addStrings(subList);
      subList.Free;
      end;//IF
    FindResult := FindNext(DSearchRec);
    end; //WHILE
  finally
    SysUtils.FindClose(FSearchRec);
  end;//TRY.. FINALLY
end;
*)

function findRegEntry(basePath, dllName: String; conduitType: TConduitType = ctStandard): String;
var
  ii: Integer;
  reg: TRegistry;
  keyNames: TStringList;
begin
  keyNames := TStringList.Create;
  reg:=TRegistry.Create;
  reg.RootKey := HKEY_CURRENT_USER;
  if not reg.OpenKey(addBackslashes(basePath), false) then raise EConduitRegistryError.Create('Problem Finding Registry Key');
  reg.getKeyNames(keyNames);
  for ii := 0 to keyNames.Count - 1 do    // Iterate
    begin
    if lowercase(copy(keyNames[ii],1, 11)) = 'application' then
      begin
      reg.OpenKey(addBackslashes(basePath) + keyNames[ii], false);
      if conduitType = ctJava then
        begin
        if lowercase(extractFileName(reg.ReadString('ClassName'))) = lowercase(extractFileName(dllName)) then
          begin
          result :=  addBackslashes(basePath) + keyNames[ii];
          break;
          end; // if lowercase(reg.ReadString('conduit')) = lowercase(dllName)...
        end
      else if conduitType = ctCOM then
        begin
        if lowercase(extractFileName(reg.ReadString('COMClient'))) = lowercase(extractFileName(dllName)) then
          begin
          result :=  addBackslashes(basePath) + keyNames[ii];
          break;
          end; // if lowercase(reg.ReadString('conduit')) = lowercase(dllName)...
        end
      else
        begin
        if (lowercase(extractFileName(reg.ReadString('conduit'))) = lowercase(extractFileName(dllName))) or
           (lowercase(extractFileName(reg.ReadString('COMClient'))) = lowercase(extractFileName(dllName))) or
           (lowercase(extractFileName(reg.ReadString('ClassName'))) = lowercase(extractFileName(dllName))) then
          begin
          result :=  addBackslashes(basePath) + keyNames[ii];
          break;
          end; // if lowercase(reg.ReadString('conduit')) = lowercase(dllName)...
        end; //if isJava.. else
      end; // if keyNames[ii] = 'application'...
    end;    // for
end;

procedure GetFileVersion(FileName: string; var Major1, Major2,
         Minor1, Minor2: Integer);
var
  Info: Pointer;
  InfoSize: DWORD;
  FileInfo: PVSFixedFileInfo;
  FileInfoSize: DWORD;
  Tmp: DWORD;
begin
  //initialize values to 0
  Major1 := 0;
  Major2 := 0;
  Minor1 := 0;
  Minor2 := 0;
  // Get the size of the FileVersionInformatioin
  InfoSize := GetFileVersionInfoSize(PChar(FileName), Tmp);
  // If InfoSize = 0, then the file may not exist, or
  // it may not have file version information in it.
  if InfoSize = 0 then
    raise Exception.Create('Can''t get file version information for '
      + FileName);
  // Allocate memory for the file version information
  GetMem(Info, InfoSize);
  try
    // Get the information
    GetFileVersionInfo(PChar(FileName), 0, InfoSize, Info);
    // Query the information for the version
    VerQueryValue(Info, '\', Pointer(FileInfo), FileInfoSize);
    // Now fill in the version information
    Major1 := FileInfo.dwFileVersionMS shr 16;
    Major2 := FileInfo.dwFileVersionMS and $FFFF;
    Minor1 := FileInfo.dwFileVersionLS shr 16;
    Minor2 := FileInfo.dwFileVersionLS and $FFFF;
  finally
    FreeMem(Info, FileInfoSize);
  end;
end;

function getFileVersionString(FileName: String): String;
var
  m1,m2,m3,m4:integer;
begin
  getfileVersion(FileName,M1,M2,M3,M4);
  result := intToStr(m1)+'.'+intToStr(m2)+'.'+intToStr(m3)+'.'+intToStr(m4);
end;

function getFirstExistingRoot(directoryName: string): String;
begin
  result := directoryName;
  while not directoryExists(result) do
    begin
    if result[length(result)] = '\' then result := copy(result, 0, length(result)-1);
    result := extractFilePath(result);
    end;    // while
end;

function InstallFile(Source, Dest: String): TIOResult;
var
  Attr: Integer;
begin
  try
    if (not FileExists(Dest)) or (not IsFileInUse(Dest)) then
      begin
      FileCopy(Source, Dest);
      //remove read only attibute, in case installing from cdrom, etc
      Attr := FileGetAttr(Dest);
      Attr := Attr and not(faReadOnly);
      FileSetAttr(Dest, Attr);
      result := ioSuccess;
      end //if not FileExists(Dest) then
    else
      begin
      if Win32Platform = VER_PLATFORM_WIN32_NT Then
         MoveFileEx( Pchar(source), Pchar(dest), MOVEFILE_REPLACE_EXISTING or MOVEFILE_DELAY_UNTIL_REBOOT)
      else
        CopyInUseSysFile(Source, Dest);
      result := ioRebootRequired;
      end; //if not FileExists(Dest).. else
  except
    result := ioFailed;
  end;
end; //procedure InstallFile(Source, Dest: String);

function IsFileInUse(FileName : string) : boolean;
var
   HFileRes : HFILE;
begin
   Result := false;
   if not FileExists(FileName) then
      exit;
   HFileRes := CreateFile(pchar(FileName), GENERIC_READ or GENERIC_WRITE, 
               0 {this is the trick!}, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
   Result := (HFileRes = INVALID_HANDLE_VALUE);
   if not Result then
   CloseHandle(HFileRes);
end;

function isNT: boolean;
var
  OSVersionInfo:TOSVersionInfo;
begin
  OSVersionInfo.dwOSVersionInfoSize:=SizeOf(OSVersionInfo);

  GetVersionEx(OSVersionInfo);
  if OSVersionInfo.dwPlatformId = VER_PLATFORM_WIN32_NT then
    result := true
  else
    result := false;
end;

function isDirectory(dirName: String): boolean;
var
  fileAttr: Integer;
begin
  fileAttr := FileGetAttr(dirName);
  if fileAttr < 0 then
    result := false
  else
    result := fileAttr and faDirectory <> 0;
end;

function KillFile(fileName: String): TIOResult;
begin
  result := ioSuccess;
  if not FileExists(fileName) then
    begin
    exit;
    end;
  try
    if not DeleteFile(PChar(FileName)) then
      begin
      if Win32Platform = VER_PLATFORM_WIN32_NT Then
         MoveFileEx(Pchar(fileName), nil, MOVEFILE_DELAY_UNTIL_REBOOT)
      else
        CopyInUseSysFile(fileName, 'NUL');
      result := ioRebootRequired;
      end; //if not FileExists(Dest).. else
  except
    result := ioFailed;
  end;
end; //procedure InstallFile(Source, Dest: String);

procedure KillDirectory(dirName:string);
var
  optStruct:TSHFileOpStruct;
begin
  try
  FillChar(optStruct, Sizeof(optStruct),#0);
  with optStruct do
    begin
      Wnd := 0;
      wFunc:=FO_DELETE;
      pFrom:=pChar(dirName);
      fFlags:=FOF_SILENT or FOF_NOCONFIRMATION;
    end;

  if (SHFileOperation(optStruct) <> 0) then
    begin
    RemoveDir(dirName);
    end; // if SHFileOperation(optStruct) <>> 0...

          //  if dirName[length(dirName)] = '\' then
  finally

  end;
end;


function ProcessParam(Param: String): String;
var
  newParam: String;
begin
  newParam:=(Param);
  if (pos('-', newParam)= 1) or (pos('/', newParam)= 1) then
    begin
    delete(newParam, 1, 1);
    end;
  result:=newParam;
end;

function ShellExec(op, fn, par, dir: PChar; show: Word; wait: Boolean): LongInt;
var
  OK: Boolean;
  Info: TShellExecuteInfo;
begin
  FillChar(Info, SizeOf(Info), Chr(0));
  Info.cbSize := SizeOf(Info);
  Info.fMask := SEE_MASK_NOCLOSEPROCESS;
  Info.lpVerb := op;
  Info.lpFile := fn;
  Info.lpParameters := par;
  Info.lpDirectory := dir;
  Info.nShow := show;
  OK := Boolean(ShellExecuteEx(@Info));
  if OK then
    begin
    //wait for the process to become idle (started)
    //new in 1.2.1 - thanks to Charles Henry for pointing this one out
    WaitForInputIdle(Info.hProcess,10000);
    if wait then
      begin
      while WaitForSingleObject(Info.hProcess,100) = WAIT_TIMEOUT do
        Application.ProcessMessages;
      OK := GetExitCodeProcess(Info.hProcess, DWord(Result));
      end
    else
      Result := 0;
    end;
  if not OK then Result := -1;
end;

function stringToConduitType(input: string): TConduitType;
begin
  if input = 'Standard Conduit' then result := ctStandard
  else if input = 'Java Conduit' then result := ctJava
  else if input = 'COM Conduit' then result := ctCom
  else
    raise Exception.Create('Invalid installation type');
end;

function conduitTypeToString(input: TConduitType): string;
begin
  case input of    //
    ctStandard: result := 'Standard Conduit';
    ctJava: result := 'Java Conduit';
    ctCom: result := 'COM Conduit';
    else
      raise Exception.Create('Invalid installation type');
    end;    // case
end;

end.
