unit installationWizardForm;

{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

{ NOTE : In a future version, We'll allow a file group to hide the individual files in it }

interface

uses
  Windows, StdCtrls, Forms, Graphics, hxInstallUtilities, SysUtils,
  Controls, ExtCtrls, Classes, CheckLst, ComCtrls, pInstallerTypes, Menus, ImgList,
  ShellCtrls, Buttons, FileCtrl, XPMan;


type
  TNodeInfo = class(TObject)
    fileIndex: integer;
    cardIndex: integer;
  end;

  TInstallationPage = (ipIntroduction, ipPreInstallMessage, ipDirectorySelection, ipFileGroups,
                       ipUserSelection, ipShortCutDirectorySelection, ipProgress,
                       ipPostInstallMessage, ipFinish, ipClose);
  TfrmInstallationWizard = class(TForm)
    btnCancel: TButton;
    btnNext: TButton;
    btnBack: TButton;
    pnlWizard: TPanel;
    pnlIntroContainer: TPanel;
    Image1: TImage;
    pnlUsers: TPanel;
    Image3: TImage;
    clbUserName: TCheckListBox;
    pnlMessage: TPanel;
    pnlLicenseAgreement: TPanel;
    rdoAccept: TRadioButton;
    rdoDecline: TRadioButton;
    rtfMessage: TRichEdit;
    pnlFileGroups: TPanel;
    tvwFileGroups: TTreeView;
    imlInstallSelection: TImageList;
    popInstallSelection: TPopupMenu;
    installToPC: TMenuItem;
    NoInstall: TMenuItem;
    Label1: TLabel;
    pnlHeader: TPanel;
    Image2: TImage;
    lblTitle: TLabel;
    lblQuestion: TLabel;
    lblInstallFor: TLabel;
    InstallToPalm: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    cbxCurrentUser: TComboBox;
    lblCurrentUser: TLabel;
    lblIntroTitle: TLabel;
    lblIntroDetails: TLabel;
    cbxUseForAll: TCheckBox;
    pnlDirectorySelection: TPanel;
    lblInstallationDirectory: TLabel;
    Image4: TImage;
    lblInstallTo: TLabel;
    btnChangeDirectory: TButton;
    pnlDirectoryDetails: TPanel;
    edtInstallationDirectory: TEdit;
    btnLookup: TButton;
    pnlShortcutDirectorySelection: TPanel;
    edtShortcutDirectory: TEdit;
    lblShortcuts: TLabel;
    lbxDirectoryNames: TListBox;
    pnlProgress: TPanel;
    pbrInstallation: TProgressBar;
    Label2: TLabel;
    lblStatus: TLabel;
    popInstallGroupSelection: TPopupMenu;
    miInstallGroup: TMenuItem;
    MenuItem2: TMenuItem;
    miInstallGroupToPalm: TMenuItem;
    MenuItem4: TMenuItem;
    miDoNotInstallGroup: TMenuItem;
    imlGroupInstallSelection: TImageList;
    XPManifest1: TXPManifest;
    procedure btnNextClick(Sender: TObject);
    procedure tvwFileGroupsGetSelectedIndex(Sender: TObject;
      Node: TTreeNode);
    procedure tvwFileGroupsMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure installToClick(Sender: TObject);
    procedure groupInstallToClick(Sender: TObject);
    procedure cbxCurrentUserChange(Sender: TObject);
    procedure btnChangeDirectoryClick(Sender: TObject);
    procedure btnLookupClick(Sender: TObject);
    procedure edtInstallationDirectoryChange(Sender: TObject);
    procedure btnBackClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure tvwFileGroupsExpanding(Sender: TObject; Node: TTreeNode;
      var AllowExpansion: Boolean);
    procedure lbxDirectoryNamesClick(Sender: TObject);
  private
    { Private declarations }
    procedure jumpToIntroduction;
    procedure jumpToPreInstall;
    procedure jumpToPostInstall;
    procedure jumpToUserSelection;
    procedure jumpToFileGroups;
    procedure jumpToFinish;
    procedure jumpToDirectorySelection;
    procedure jumpToShortCutDirectorySelection;
    function allUsersHaveSelectedCards: boolean;
    procedure applySettingsToRemainingUsers;
    function validateFileGroups: boolean;
    function validateCurrentPage(closeOnInvalid: boolean=false; closeMessage: string=''): boolean;
    function getNextPage: TInstallationPage;
    function getPreviousPage: TInstallationPage;
    function getNextAfterIntro:TInstallationPage;
    function getNextAfterDirectorySelect: TInstallationPage;
    function getNextAfterPreInstallationMessage:TInstallationPage;
    function getNextAfterFileGroups:TInstallationPage;
    function getNextAfterUserSelection:TInstallationPage;
    function getNextAfterShortcutDirectory: TInstallationPage;
    function getNextAfterProgress:TInstallationPage;
    function getNextAfterPostInstallMessage:TInstallationPage;
    function getPreviousBeforeDirectorySelect: TInstallationPage;
    function getPreviousBeforePreInstallationMessage:TInstallationPage;
    function getPreviousBeforeFileGroups:TInstallationPage;
    function getPreviousBeforeUserSelection:TInstallationPage;
    function getPreviousBeforeShortcutDirectory: TInstallationPage;
    function groupIsAllPalm: boolean;
    function groupIsAllCardEnabled: boolean;
    procedure HidePanels;
  public
    { Public declarations }
    currentPage: TInstallationPage;
    installerSettings: TPalmInstaller;
    procedure selectAllUsers;
    procedure jumpToProgress;
    procedure jumpToPage(page: TInstallationPage);
    constructor create(AOwner: TComponent); override;
  end;

var
  frmInstallationWizard: TfrmInstallationWizard;

implementation

{$R *.dfm}

constructor TfrmInstallationWizard.create(AOwner: TComponent);
begin
  inherited;
  currentPage := ipIntroduction;
end;

procedure TfrmInstallationWizard.HidePanels;
begin
  pnlIntroContainer.Visible := false;
  pnlUsers.Visible := false;
  pnlLicenseAgreement.Visible := false;
  pnlMessage.visible := false;
  pnlUsers.Visible := false;
  pnlFileGroups.Visible := false;
  pnlShortcutDirectorySelection.Visible := false;
  pnlDirectorySelection.Visible:=false;
  pnlProgress.Visible := false;
end;

procedure TfrmInstallationWizard.jumpToShortCutDirectorySelection;
var
  startMenu: string;
  DSearchRec: TSearchRec;
  FindResult: integer;
begin
  HidePanels;
  lbxDirectoryNames.Clear;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  btnBack.Enabled := true;
  startMenu := IncludeTrailingPathDelimiter(extractFileName(ExcludeTrailingPathDelimiter(extractFilePath(getStartProgramsMenuPath)))) +
               IncludeTrailingPathDelimiter(extractFileName(getStartProgramsMenuPath));
  lblShortcuts.Caption := 'Create Shortcuts in ' + startMenu + ':';
  
  FindResult := FindFirst(IncludeTrailingPathDelimiter(getStartProgramsMenuPath) + '*.*' ,faDirectory, DSearchRec);
  try
    { Continue to search for the files according to the specified
      mask. If found, add the files and their paths to the stringlist.}
    while (FindResult = 0) do
      begin
      if ((DSearchRec.Attr and faDirectory)=faDirectory) and
                (DSearchRec.Name <> '..') and (DSearchRec.Name <> '.') then
        lbxDirectoryNames.Items.Add(DSearchRec.Name);
      FindResult := FindNext(DSearchRec);
      end; //WHILE
  finally
   SysUtils.FindClose(DSearchRec);
  end;//TRY.. FINALLY
  edtShortcutDirectory.Text := installerSettings.shortCutPath;
  pnlShortcutDirectorySelection.Visible := true;
  lblTitle.Caption := 'Select a Shortcut Location';
  lblQuestion.Caption := 'Please select the folder on the Start Menu where you would like to ' +
                         'create shortcuts for' + installerSettings.productName +'.';
end; //

procedure TfrmInstallationWizard.jumpToIntroduction;
begin
  HidePanels;
  pnlWizard.Visible := false;
  pnlIntroContainer.Visible := true;
  pnlIntroContainer.BringToFront;
  lblIntroTitle.Caption := 'Welcome to the handX pInstaller Wizard for ' + installerSettings.productName;
  lblIntroDetails.Caption := 'This wizard will walk you through the steps necessary to install ' +
                             installerSettings.productName;
  btnBack.Enabled := false;
end; //

procedure TfrmInstallationWizard.jumpToDirectorySelection;
begin
  HidePanels;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  btnBack.Enabled := true;

  lblInstallationDirectory.Caption := installerSettings.installationPath;
  lblInstallTo.Caption := 'Install ' + installerSettings.productName + ' to:';
  pnlDirectorySelection.Visible := true;
  lblTitle.Caption := 'Select a Folder';
  lblQuestion.Caption := 'Please select the folder where you would like to install ' +
                         installerSettings.productName +'.';
end;    //

procedure TfrmInstallationWizard.jumpToPreInstall;
begin
  HidePanels;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  pnlMessage.visible := true;
  btnBack.Enabled := true;
  pnlLicenseAgreement.Visible := installerSettings.preInstallShowLicense;
  if (pnlLicenseAgreement.Visible) and (installerSettings.preInstallSetDefault) then
    begin
    rdoAccept.Checked := installerSettings.preInstallDefaultAs;
    rdoDecline.Checked := not installerSettings.preInstallDefaultAs;
    end;
  lblQuestion.Caption := installerSettings.preInstallMsgCaption;
  lblTitle.Caption := installerSettings.preInstallMsgTitle;
  self.rtfMessage.Lines.LoadFromFile(installerSettings.preInstallMsgFile);
  pnlWizard.BringToFront;
end;    //

procedure TfrmInstallationWizard.jumpToPostInstall;
begin
//  show;
  if installerSettings.runSilent and installerSettings.postInstallationMessageOverrideSilentMode then
    begin
    if (installerSettings.preInstallationMessageOverrideSilentMode) or
                          (installerSettings.userPromptOverrideSilentMode) then
      show;
    btnNext.Visible := false;
    btnCancel.Caption := 'OK';
    end;
  pnlWizard.BringToFront;
  HidePanels;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  pnlMessage.visible := true;
  btnBack.Enabled := false;
  pnlLicenseAgreement.Visible := false;
  lblQuestion.Caption := installerSettings.postInstallMsgCaption;
  lblTitle.Caption := installerSettings.postInstallMsgTitle;
  self.rtfMessage.Lines.LoadFromFile(installerSettings.postInstallMsgFile);
end;    //

procedure TfrmInstallationWizard.jumpToUserSelection;
var
  ii: Integer;
begin
  HidePanels;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  btnBack.Enabled := true;
  pnlUsers.Visible := true;
  lblTitle.Caption := 'Select Users for Installation';
  lblQuestion.Caption := 'Check each HotSync User that should have this program installed.';
  if clbUserName.Count = 0 then
    begin
    installerSettings.AddLogEntry('starting multiple installs');
    for ii := Low(installerSettings.HotSyncUsers) to High(installerSettings.HotSyncUsers) do
      begin
      clbUserName.Items.Add(installerSettings.HotSyncUsers[ii].displayName);
      if Length(installerSettings.installUsers) = 0 then
        begin
        clbUserName.Checked[clbUserName.Count - 1] := true;
        end
      else
        begin
        clbUserName.Checked[clbUserName.Count - 1] := installerSettings.isInstalledUser(installerSettings.HotSyncUsers[ii].id);
        end;
      end; // for
    end;
end;    //

procedure TfrmInstallationWizard.jumpToFileGroups;
var
  ii: Integer;
  groupIndex: integer;
  fileIndex: integer;
  rootNode: TTreeNode;
  groupNode: TTreeNode;
  fileNode: TTreeNode;
  pcFilesInGroup : boolean;
  palmFilesInGroup : boolean;
  groupSize: extended;
begin
  HidePanels;
  pnlWizard.Visible := true;
  pnlWizard.BringToFront;
  btnBack.Enabled := true;
  pnlWizard.Visible := true;
  pnlFileGroups.Visible := true;
  if Length(installerSettings.installUsers) > 1 then
    begin
    lblCurrentUser.Visible := true;
    cbxCurrentUser.Visible := true;
    for ii := Low(installerSettings.installUsers) to High(installerSettings.installUsers) do
      cbxCurrentUser.Items.Add(installerSettings.installUsers[ii].displayName);
    cbxCurrentUser.ItemIndex := 0;
    cbxUseForAll.Visible := true;
    cbxUseForAll.Checked := true;
    end
  else
    begin
    lblCurrentUser.Visible := false;
    cbxCurrentUser.Visible := false;
    for ii := Low(installerSettings.installUsers) to High(installerSettings.installUsers) do
      cbxCurrentUser.Items.Add(installerSettings.installUsers[ii].displayName);
    cbxCurrentUser.ItemIndex := 0;
    cbxUseForAll.Visible := false;
    cbxUseForAll.Checked := false;
    end;
  cbxCurrentUserChange(cbxCurrentUser);
  if tvwFileGroups.Items.Count = 0 then
    begin
    rootNode := tvwFileGroups.Items.Add(nil, installerSettings.productName);
    for groupIndex := Low(installerSettings.FileGroups) to High(installerSettings.FileGroups) do
      begin
      //if the group has only one file, associate the groups data with the group node, and don't show the file
      groupSize := 0;
      if installerSettings.getFileGroupCount(installerSettings.FileGroups[groupIndex].id) = 1 then
        begin
        groupNode := tvwFileGroups.Items.AddChild(rootNode, installerSettings.FileGroups[groupIndex].description);
        for fileIndex := Low(installerSettings.InstallerFiles) to High(installerSettings.InstallerFiles) do
          begin
          if installerSettings.installerFiles[fileIndex].fileGroup = installerSettings.fileGroups[groupIndex].id then
            begin
            groupSize := groupSize + getFileSize(installerSettings.installerFiles[fileIndex].fileName);
            if installerSettings.fileGroups[groupIndex].installByDefault then
              begin
              if upperCase(installerSettings.installerFiles[fileIndex].destination) = '$PALM_DEVICE' then
                begin
                if installerSettings.installerFiles[fileIndex].forceCardInstallation then
                  groupNode.ImageIndex := 2
                else
                  groupNode.ImageIndex := 1;
                end //if upperCase(installerSettings.installerFiles[fileIndex].destination) = '$PALM_DEVICE' then
              else
                begin
                groupNode.ImageIndex := 0;
                end;//if upperCase(installerSettings.installerFiles[fileIndex].destination) = '$PALM_DEVICE'.. else
              end //if installerSettings.fileGroups[groupIndex].installByDefault then
            else
              begin
              groupNode.ImageIndex := 3;
              end; //if installerSettings.fileGroups[groupIndex].installByDefault.. else
            groupNode.Data := TNodeInfo.Create;
            TNodeInfo(groupNode.data).fileIndex := fileIndex;
            if not installerSettings.installerFiles[fileIndex].forceCardInstallation then
              TNodeInfo(groupNode.Data).cardIndex := -1
            else
              TNodeInfo(groupNode.Data).cardIndex := 0;  // install to first available card;

            end; //if installerSettings.installerFiles[fileIndex].fileGroup = installerSettings.fileGroups[groupIndex].id then
          end; // for fileIndex := Low(installerSettings.InstallerFiles) to High(installerSettings.InstallerFiles) do
          groupNode.Text := groupNode.Text + ' (' + fileSizeToPrintable(trunc(groupSize)) + ')';
          groupNode.SelectedIndex := -1;
          continue;
        end; //if installerSettings.getFileGroupCount(installerSettings.FileGroups[groupIndex].id) = 1

      groupNode := tvwFileGroups.Items.AddChild(rootNode, installerSettings.FileGroups[groupIndex].description);
      groupNode.SelectedIndex := -1;
      if installerSettings.fileGroups[groupIndex].installByDefault then
        begin
        groupNode.ImageIndex := 8
        end // if installerSettings.fileGroups[groupIndex].installByDefault..
      else
        groupNode.ImageIndex := 3;
      pcFilesInGroup := false;
      palmFilesInGroup := false;
      for fileIndex := Low(installerSettings.InstallerFiles) to High(installerSettings.InstallerFiles) do
        begin
        if installerSettings.installerFiles[fileIndex].fileGroup = installerSettings.fileGroups[groupIndex].id then
          begin
          groupSize := groupSize + getFileSize(installerSettings.installerFiles[fileIndex].fileName);
          fileNode := tvwFileGroups.Items.AddChild(groupNode, installerSettings.installerFiles[fileIndex].fileDescription);
          fileNode.ImageIndex := groupNode.ImageIndex;
          if upperCase(installerSettings.installerFiles[fileIndex].destination) = '$PALM_DEVICE' then
            begin
            palmFilesInGroup := true;
            if groupNode.ImageIndex <> 3 then
              begin
              if installerSettings.installerFiles[fileIndex].forceCardInstallation then
                begin
                fileNode.ImageIndex := 2;
                end
              else
                begin
                fileNode.ImageIndex := 1;
                end; //if installerSettings.installerFiles[fileIndex].forceCardInstallation.. else
              end //if groupNode.ImageIndex <> 3 then
            end
          else
            begin
            pcFilesInGroup := true;
            end;
          if not installerSettings.installerFiles[fileIndex].installByDefault then
            fileNode.ImageIndex := 3;
          fileNode.Data := TNodeInfo.Create;
          TNodeInfo(fileNode.data).fileIndex := fileIndex;
          if not installerSettings.installerFiles[fileIndex].forceCardInstallation then
            TNodeInfo(fileNode.data).cardIndex := -1
          else
            TNodeInfo(fileNode.data).cardIndex := 0
          end;
        end; // for fileIndex := Low(installerSettings.InstallerFiles) to High(installerSettings.InstallerFiles) do
      if (palmFilesInGroup) and (not pcFilesInGroup) then
        begin
        groupNode.ImageIndex := 1;
        end;
      if groupNode.ImageIndex < 3 then
        groupNode.ImageIndex := groupNode.ImageIndex + 8;
      groupNode.Text := groupNode.Text + ' (' + fileSizeToPrintable(trunc(groupSize)) + ')';
      end; // for

    rootNode.Expand(false);
    end;
  lblTitle.Caption := 'Installation Options for ' + installerSettings.productName;
  lblQuestion.Caption := 'Select the components that you wish to install';
end;    //

procedure TfrmInstallationWizard.jumpToProgress;
var
  fileIndex: Integer;
  userIndex: Integer;
  installedFiles: TStringList;
  currentFile : TPalmInstallerFile;
begin
  btnNext.Enabled := false;
  btnBack.Enabled := false;
  HidePanels;
  pnlProgress.Visible := true;
  pbrInstallation.Max := Length(installerSettings.installerFiles);
  installedFiles := TStringList.create;
  lblTitle.Caption := 'Installation Progress';
  lblQuestion.Caption := 'Now installing ' + installerSettings.productName +'.';
  for fileIndex := Low(installerSettings.installerFiles) to High(installerSettings.installerFiles) do
    begin
    Application.ProcessMessages;
    currentFile := installerSettings.installerFiles[fileIndex];
    if currentFile.doNotInstall then
      begin
      pbrInstallation.StepIt;
      continue;
      end;
    if (currentFile.fileGroup = 0) or (installerSettings.fileIsSelectedForInstallation(fileIndex)) then
      begin
      lblStatus.Caption := 'Installing ' + currentFile.fileName + '... ';
      Application.ProcessMessages;
      if uppercase(currentFile.destination) = '$PALM_DEVICE' then
        begin
        //we need to install it multiple times
        for userIndex := Low(installerSettings.installUsers) to High(installerSettings.installUsers) do
          begin
          if (currentFile.Filegroup = 0) or (installerSettings.installUsers[userIndex].fileSelectedForInstall(fileIndex)) then
            begin
            lblStatus.Caption := 'Installing ' + currentFile.fileName + ' for ' +
                 installerSettings.installUsers[userIndex].displayName + '... ';
            Application.ProcessMessages;
            installerSettings.installFile(fileIndex, userIndex);
            //install for that user
            end;
          end;
        end
      else
        begin
        if installedFiles.IndexOf(installerSettings.installerFiles[fileIndex].fileName) = -1 then
          begin
          installerSettings.installFile(fileIndex, 0);
          installedFiles.Add(installerSettings.installerFiles[fileIndex].fileName)
          end;
        end;
      end;
    pbrInstallation.StepIt;
    end;
  installedFiles.Free;
  //now add registry settings
  lblStatus.Caption := 'Applying registry settings... ';
  installerSettings.applyCustomRegistryChanges;
  if installerSettings.ConduitSelectedForInstallation then
    begin
    lblStatus.Caption := 'Restarting HotSync';
    installerSettings.restartHotSync;
    end;
  //now run any post install executables
  lblStatus.Caption := 'Applying custom installation additions... ';
  installerSettings.runPostInstallExes;
  installerSettings.buildUninstaller;
  lblStatus.Caption := 'Installation Complete';
  if installerSettings.runSilent then
    begin
    if not installerSettings.postInstallationMessageOverrideSilentMode then
      begin
      close;
      exit;
      end;
    end;
  btnNext.Enabled := true;
  jumpToPage(getNextPage);
end;

procedure TfrmInstallationWizard.jumpToFinish;
begin
  btnNext.Caption := 'Finish';
  btnCancel.Enabled := false;
  btnNext.Enabled := true;
  pnlIntroContainer.BringToFront;
  pnlIntroContainer.Visible := true;
  lblIntroTitle.Caption := 'Installation Complete.';
  lblIntroDetails.Caption := installerSettings.productName + ' is now installed.';
end;    //

procedure TfrmInstallationWizard.jumpToPage(page: TInstallationPage);
begin
  LockWindowUpdate(Handle);
// set the currentPage to
  currentPage:= page;
  if page = ipIntroduction then
    jumpToIntroduction
  else if page = ipPreInstallMessage then
    jumpToPreInstall
  else if page = ipDirectorySelection then
    jumpToDirectorySelection
  else if page = ipUserSelection then
     jumpToUserSelection
  else if page = ipFileGroups then
    jumpToFileGroups
  else if page = ipShortCutDirectorySelection then
    jumpToShortCutDirectorySelection
  else if page = ipProgress then
    begin
    LockWindowUpdate(0);
    jumpToProgress;
    end
  else if page = ipPostInstallMessage then
    jumpToPostInstall
  else if page = ipFinish then
    jumpToFinish
  else if page = ipClose then
    close;
  LockWindowUpdate(0);
end;

function TfrmInstallationWizard.allUsersHaveSelectedCards: boolean;
var
  ii: integer;
  cardUsedList: TStringList;
  cardName: string;
begin
  cardUsedList := TStringList.Create;
  for ii := Low(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix) to
          High(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix) do
    begin
    if installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix[ii] <> -1 then
      begin
      cardName := installerSettings.getSlotName(installerSettings.installUsers[cbxCurrentUser.ItemIndex].id,
             installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix[ii]);
      if cardUsedList.IndexOf(cardName) = -1 then
        cardUsedList.Add(cardName);
      end;
    end; // for
  result := true;
  for ii := 0 to cardUsedList.Count - 1 do
    begin
    if not installerSettings.allUsersHaveMatchingCard(cardUsedList[ii]) then
      begin
      result := false;
      break;
      end;
    end; // for
  cardUsedList.Free;
end;

procedure TfrmInstallationWizard.applySettingsToRemainingUsers;
var
  ii, jj: Integer;
  cardName: string;
begin
  for ii := cbxCurrentUser.ItemIndex + 1 to cbxCurrentUser.Items.Count - 1 do
    begin
    installerSettings.installUsers[ii].installFileMatrix :=
                    installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileMatrix;
    installerSettings.installUsers[ii].installCardFileMatrix :=
                    installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix;
    for jj := Low(installerSettings.installUsers[ii].installCardFileMatrix) to
                                High(installerSettings.installUsers[ii].installCardFileMatrix) do
      begin
      //if they have no cards we set the installation to the palm
      if installerSettings.getInstallableSlotCount(installerSettings.installUsers[ii].id) = 0 then
        installerSettings.installUsers[ii].installCardFileMatrix[jj] := -1
      else
        begin
        if installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix[jj] <> -1 then
          begin
          //if they don't have a matching card, then we install it to the first available card (0)
          cardName := installerSettings.getSlotName(installerSettings.installUsers[cbxCurrentUser.ItemIndex].id,
                 installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix[jj]);
          if not installerSettings.userHasMatchingCard(installerSettings.installUsers[cbxCurrentUser.ItemIndex].id, cardName) then
            begin
            installerSettings.installUsers[ii].installCardFileMatrix[jj] := 0;
            end;
          end; //
        end;
      end; //for jj
    end; // for
end;

function TfrmInstallationWizard.validateFileGroups: boolean;
var
  ii: integer;
  cardUsed: boolean;
  missingCards: boolean;
  mismatchedCards: boolean;
  continueInstallation: boolean;
begin
  cardUsed := false;
  result := false;
{  setLength(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileIndexes, 0);
  setLength(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileIndexes, 0);
}
  for ii := 0 to tvwFileGroups.Items.Count - 1 do
    begin
    if tvwFileGroups.Items[ii].Data = nil then continue;
    if tvwFileGroups.Items[ii].ImageIndex <> 3 then
      begin
      installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileMatrix
                            [TNodeInfo(tvwFileGroups.Items[ii].Data).fileIndex] := 1;
      installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileMatrix
                            [TNodeInfo(tvwFileGroups.Items[ii].Data).fileIndex] :=
                            TNodeInfo(tvwFileGroups.Items[ii].Data).cardIndex;
      cardUsed := TNodeInfo(tvwFileGroups.Items[ii].Data).cardIndex <> -1;

{      currentLength := Length(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileIndexes);
      setLength(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileIndexes,
                    currentLength + 1);
      installerSettings.installUsers[cbxCurrentUser.ItemIndex].installFileIndexes[currentLength] :=
                     TNodeInfo(tvwFileGroups.Items[ii].Data).fileIndex;
      setLength(installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileIndexes,
                    currentLength + 1);
      installerSettings.installUsers[cbxCurrentUser.ItemIndex].installCardFileIndexes[currentLength] :=
              TNodeInfo(tvwFileGroups.Items[ii].Data).cardIndex;
      cardUsed := TNodeInfo(tvwFileGroups.Items[ii].Data).cardIndex <> -1;
}
      end;
    end;
  if not cbxUseForAll.Checked then
    begin
    if (cbxCurrentUser.ItemIndex + 1 < cbxCurrentUser.Items.Count) then
      begin
      cbxCurrentUser.ItemIndex := cbxCurrentUser.ItemIndex + 1;
      cbxCurrentUserChange(cbxCurrentUser);
      exit;
      end
    else
      begin
      result := true;
      exit;
      end;
    end;

  //if the card was selected and there are more than one users who are supposed to use the same
  //settings we need to make sure they all have cards with the same name or we have to make
  //some other assumptions
  if (cardUsed) and (cbxUseForAll.Checked) and (cbxCurrentUser.Items.Count > 1) and
                        (cbxCurrentUser.ItemIndex + 1 < cbxCurrentUser.Items.Count) then
    begin
    //check to verify that all users have installable cards
    missingCards := not installerSettings.allUsersHaveCards;
    //check to see that all users have a matching card
    mismatchedCards := not allUsersHaveSelectedCards;

    if missingCards and mismatchedCards then
      continueInstallation := Application.MessageBox('Not all users selected for installation have ' +
                              'installable expansion cards'+Char(10)+'and not all selected '+
                              'expansion cards are available for all users with' + Char(10) +
                              'expansion cards.' + Char(10) + Char(10) + 'Do you want ' +
                              'handX pInstaller Wizard to install directly to the ' +
                              'Palm' + Char(10) + 'device, or the first installable expansion card, ' +
                              'for those users?',
                              'handX pInstaller Wizard', mb_yesno or mb_iconquestion) = idYes
    else if missingCards then
      continueInstallation := Application.MessageBox('Not all users selected for installation have ' +
                              'installable expansion cards.'+Char(10)+Char(10)+'Do you want ' +
                              'handX pInstaller Wizard to install directly to the ' +
                              'Palm'+Char(10)+'device for those users?',
                              'handX pInstaller Wizard ', mb_yesno or mb_iconquestion) = idYes
    else if mismatchedCards then
      continueInstallation := Application.MessageBox('Not all users selected for installation have ' +
                              'a matching expansion card.' + Char(10)+ Char(10) + 'Do you want ' +
                              'handX pInstaller Wizard to install directly to the ' +
                              'first' + Char(10) + 'installable expansion card for those users?',
                              'handX pInstaller Wizard', mb_yesno or mb_iconquestion) = idYes
    else
      begin
      continueInstallation := true;
      end;
    end
  else
    begin
    continueInstallation := true;
    end;

  if continueInstallation then
    begin
    applySettingsToRemainingUsers;
    result := true;
    end
  else
    begin
    cbxUseForAll.Checked := false;
    exit;
    end;
end;    // function TfrmInstallationWizard.valiadeUserPage: boolean;

procedure TfrmInstallationWizard.selectAllUsers;
var
  ii: integer;
begin
  SetLength(installerSettings.installUsers, 0);
  for ii := low(installerSettings.hotSyncUsers) to high(installerSettings.hotSyncUsers) do    // Iterate
    begin
    setLength(installerSettings.installUsers, Length(installerSettings.installUsers)+1);
    installerSettings.installUsers[Length(installerSettings.installUsers)-1] :=
                       TInstallerUser.create(installerSettings.hotSyncUsers[ii].index,
                       installerSettings.hotSyncUsers[ii].id,
                       installerSettings.hotSyncUsers[ii].displayName);
    installerSettings.installUsers[Length(installerSettings.installUsers)-1].initializeFileMatrixes(installerSettings.installerFiles);
    end;    // for
end;

function TfrmInstallationWizard.validateCurrentPage(closeOnInvalid: boolean=false; closeMessage: string=''): boolean;
var
  ii: Integer;
begin
  result := false;
  if currentPage = ipIntroduction then result := true
  else if currentPage = ipPreInstallMessage then
    begin
    if not pnlLicenseAgreement.visible then result := true
    else result := rdoAccept.Checked;
    end
  else if currentPage = ipDirectorySelection then
    begin
    result := directoryExists(extractFileDrive(lblInstallationDirectory.Caption));
    installerSettings.installationPath := lblInstallationDirectory.Caption;
    end
  else if currentPage = ipShortCutDirectorySelection then
    begin
    if (pos('/', edtShortcutDirectory.text) <> 0) or
                 (pos('/', edtShortcutDirectory.text) <> 0) or
                 (pos(':', edtShortcutDirectory.text) <> 0) or
                 (pos('?', edtShortcutDirectory.text) <> 0) or
                 (pos('<', edtShortcutDirectory.text) <> 0) or
                 (pos('>', edtShortcutDirectory.text) <> 0) or
                 (pos('|', edtShortcutDirectory.text) <> 0) or
                 (pos('"', edtShortcutDirectory.text) <> 0) then
      begin
      Application.MessageBox('Your Shorcut Path cannot contain invalid characters (/ : * ? < > | ").',
                           'Shortcut Path not valid', MB_ICONSTOP);
      result := false;
      end
    else
      begin
      installerSettings.shortCutPath := edtShortcutDirectory.Text;
      result := true;
      end; //if (pos('/' . .else
    end
  else if currentPage = ipUserSelection then
    begin
    SetLength(installerSettings.installUsers, 0);
    for ii := 0 to clbUserName.Count - 1 do    // Iterate
      begin
      if clbUserName.Checked[ii] then
        begin
        setLength(installerSettings.installUsers, Length(installerSettings.installUsers)+1);
        installerSettings.installUsers[Length(installerSettings.installUsers)-1] :=
                           TInstallerUser.create(installerSettings.hotSyncUsers[ii].index,
                           installerSettings.hotSyncUsers[ii].id,
                           installerSettings.hotSyncUsers[ii].displayName);
        installerSettings.installUsers[Length(installerSettings.installUsers)-1].initializeFileMatrixes(installerSettings.installerFiles);
        result := true;
        end;
      end;    // for
    end
  else if currentPage = ipFileGroups then
    begin
    result := validateFileGroups;
    end
  else if currentPage = ipPostInstallMessage then
    begin
    result := true;
    end
  else
    begin

    end;
end;

function TfrmInstallationWizard.getNextAfterIntro:TInstallationPage;
begin
  if installerSettings.preInstallMsgFile <> '' then
    begin
    result := ipPreInstallMessage;
    exit;
    end
  else
    begin
    result := getNextAfterPreInstallationMessage;
    end;
end;

function TfrmInstallationWizard.getNextAfterPreInstallationMessage:TInstallationPage;
begin
  if installerSettings.runSilent then
    begin
    if installerSettings.userPromptOverrideSilentMode then
      result := ipUserSelection
    else
      begin
      selectAllUsers;
      hide;
      result := ipProgress;
      end; //if installerSettings.userPromptOverrideSilentMode then
    exit;
    end;
  if (installerSettings.installationPath <> '') and (installerSettings.allowUserPathSelection) then
    result := ipDirectorySelection
  else result := getNextAfterDirectorySelect;
end;

function TfrmInstallationWizard.getNextAfterDirectorySelect: TInstallationPage;
begin
  if installerSettings.allowUserShortcutPathPrompt then
    result := ipShortCutDirectorySelection
  else
    result := getNextAfterShortcutDirectory;
end;

function TfrmInstallationWizard.getNextAfterUserSelection:TInstallationPage;
begin
  if installerSettings.runSilent then
    begin
    hide;
    result := ipProgress;
    exit;
    end; //if installerSettings.userPromptOverrideSilentMode then
  if Length(installerSettings.fileGroups) > 0 then
    begin
    result := ipFileGroups;
    end
  else
    result := getNextAfterFileGroups;
end;

function TfrmInstallationWizard.getNextAfterFileGroups:TInstallationPage;
begin
  result := ipProgress;
end;

function TfrmInstallationWizard.getNextAfterProgress:TInstallationPage;
begin
  if installerSettings.postInstallMsgFile <> '' then
    result := ipPostInstallMessage
  else
  result := getNextAfterPostInstallMessage;
end;

function TfrmInstallationWizard.getNextAfterPostInstallMessage:TInstallationPage;
begin
  if installerSettings.runSilent then
    result := ipClose
  else
    result := ipFinish;
end;


function TfrmInstallationWizard.getNextAfterShortcutDirectory: TInstallationPage;
begin
  if installerSettings.palmFilesToInstall then
    begin
    if installerSettings.GetUserCount > 1 then
      begin
      result := ipUserSelection;
      exit;
      end
    else
      begin
      installerSettings.AddLogEntry('one user');
      setLength(installerSettings.installUsers, Length(installerSettings.installUsers)+1);
      installerSettings.installUsers[Length(installerSettings.installUsers)-1] :=
                           TInstallerUser.create(installerSettings.hotSyncUsers[0].index,
                           installerSettings.hotSyncUsers[0].id,
                           installerSettings.hotSyncUsers[0].displayName);
      installerSettings.installUsers[Length(installerSettings.installUsers)-1].initializeFileMatrixes(installerSettings.installerFiles);
      end;
    end;
  result := getNextAfterUserSelection;
end;

function TfrmInstallationWizard.getNextPage: TInstallationPage;
begin
  case currentPage of    //
    ipIntroduction: result := getNextAfterIntro;
    ipPreInstallMessage : result := getNextAfterPreInstallationMessage;
    ipDirectorySelection : result := getNextAfterDirectorySelect;
    ipShortCutDirectorySelection : result := getNextAfterShortcutDirectory;
    ipUserSelection     : result := getNextAfterUserSelection;
    ipFileGroups        : result := getNextAfterFileGroups;
    ipProgress          : result := getNextAfterProgress;
    ipPostInstallMessage: result := getNextAfterPostInstallMessage;
  else
    result := ipFinish;
  end;    // case
end;

function TfrmInstallationWizard.getPreviousBeforeFileGroups:TInstallationPage;
begin
    if (cbxCurrentUser.ItemIndex > 0) then
      begin
      cbxCurrentUser.ItemIndex := cbxCurrentUser.ItemIndex - 1;
      cbxCurrentUserChange(cbxCurrentUser);
      result := ipFileGroups;
      exit;
      end;

  if installerSettings.palmFilesToInstall then
    begin
    if installerSettings.GetUserCount > 1 then
      begin
      result := ipUserSelection;
      exit;
      end;
    end;
  result := getPreviousBeforeUserSelection;
end;

function TfrmInstallationWizard.getPreviousBeforeUserSelection:TInstallationPage;
begin
  if installerSettings.allowUserShortcutPathPrompt then
    result := ipShortCutDirectorySelection
  else
    result := getPreviousBeforeShortcutDirectory;
end;


function TfrmInstallationWizard.getPreviousBeforeShortcutDirectory: TInstallationPage;
begin
  if (installerSettings.installationPath <> '') and (installerSettings.allowUserPathSelection) then
    result := ipDirectorySelection
  else
    result := getPreviousBeforeDirectorySelect;
end;

function TfrmInstallationWizard.getPreviousBeforeDirectorySelect: TInstallationPage;
begin
  if installerSettings.preInstallMsgFile <> '' then
    result := ipPreInstallMessage
  else result := getPreviousBeforePreInstallationMessage;
end;

function TfrmInstallationWizard.getPreviousBeforePreInstallationMessage:TInstallationPage;
begin
  result := ipIntroduction;
end;

function TfrmInstallationWizard.getPreviousPage: TInstallationPage;
begin
  case currentPage of    //
    ipPreInstallMessage : result := ipIntroduction;
    ipDirectorySelection : result := getPreviousBeforeDirectorySelect;
    ipShortCutDirectorySelection : result := getPreviousBeforeShortcutDirectory;
    ipUserSelection     : result := getPreviousBeforeUserSelection;
    ipFileGroups        : result := getPreviousBeforeFileGroups;
  else
    result := ipFinish;
  end;    // case
end;

procedure TfrmInstallationWizard.btnNextClick(Sender: TObject);
begin
  if installerSettings = nil then
    raise Exception.Create('Installation Settings have not been loaded');
  if btnNext.Caption = 'Finish' then
    Close;
  if validateCurrentPage then
    jumpToPage(getNextPage);
end;

procedure TfrmInstallationWizard.tvwFileGroupsGetSelectedIndex(
  Sender: TObject; Node: TTreeNode);
begin
  node.SelectedIndex := node.ImageIndex;
end;

function TfrmInstallationWizard.groupIsAllCardEnabled: boolean;
var
  childNode: TTreeNode;
begin
  result := true;
  childNode := tvwFileGroups.Selected.getFirstChild;
  while childNode <> nil do
    begin
    if childNode.Data <> nil then
      begin
      if uppercase(installerSettings.installerFiles[TNodeInfo(childNode.Data).fileIndex].destination) <> '$PALM_DEVICE' then
        begin
        result := false;
        exit;
        end
      else if not installerSettings.installerFiles[TNodeInfo(childNode.Data).fileIndex].cardInstallable then
        begin
        result := false;
        exit;
        end;
      end;
    childNode := tvwFileGroups.Selected.GetNextChild(childNode);
    end;
end;

function TfrmInstallationWizard.groupIsAllPalm: boolean;
var
  childNode: TTreeNode;
begin
  result := true;
  childNode := tvwFileGroups.Selected.getFirstChild;
  while childNode <> nil do
    begin
    if childNode.Data <> nil then
      begin
      if uppercase(installerSettings.installerFiles[TNodeInfo(childNode.Data).fileIndex].destination) <> '$PALM_DEVICE' then
        begin
        result := false;
        exit;
        end;
      end;
    childNode := tvwFileGroups.Selected.GetNextChild(childNode);
    end;
end;

procedure TfrmInstallationWizard.tvwFileGroupsMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ii: Integer;
  offSet: TPoint;
  checkRect: TRect;
  isPalmFile: boolean;
  isCardInstallable: boolean;
  menuIndex: integer;
begin
  if tvwFileGroups.Selected.Level = 0 then exit;
  if tvwFileGroups.Selected.Data <> nil then
    begin
    isPalmFile := uppercase(installerSettings.installerFiles[TNodeInfo(tvwFileGroups.Selected.Data).FileIndex].destination) = '$PALM_DEVICE';
    isCardInstallable := installerSettings.installerFiles[TNodeInfo(tvwFileGroups.Selected.Data).FileIndex].cardInstallable;
    for ii := 0 to popInstallSelection.Items.Count - 1 do
      begin
      if (popInstallSelection.Items[ii] <> installToPC) and (popInstallSelection.Items[ii]<> noInstall) then
        popInstallSelection.Items[ii].Enabled := isPalmFile;
      menuIndex := 3;
      while popInstallSelection.Items[menuIndex].Caption <> '-' do
        begin
        if popInstallSelection.Items[menuIndex].Caption = '-' then
          begin
          break;
          end;
        popInstallSelection.Items[menuIndex].Enabled := isCardInstallable;
        inc(menuIndex);
        end;    // while
      end; // for
    installToPC.Enabled := not isPalmFile;
    installToPalm.Enabled := (isPalmFile) and
        (not installerSettings.installerFiles[TNodeInfo(tvwFileGroups.Selected.Data).FileIndex].forceCardInstallation);
    end
  else
    begin
    isPalmFile := groupIsAllPalm;
    isCardInstallable := groupIsAllCardEnabled;
    popInstallGroupSelection.Items[2].Enabled := isPalmFile;
    menuIndex := 3;
    while popInstallGroupSelection.Items[menuIndex].Caption <> '-' do
      begin
      if popInstallGroupSelection.Items[menuIndex].Caption = '-' then
        begin
        break;
        end;
      popInstallGroupSelection.Items[menuIndex].Enabled := isCardInstallable;
      inc(menuIndex);
      end;    // while
    miInstallGroup.Enabled := not isPalmFile;
    end; //if tvwFileGroups.Selected.Data <> nil.. else
  checkRect := tvwFileGroups.Selected.DisplayRect(true);
  checkRect.Left := checkRect.Left - 32;
  checkRect.Right := checkRect.Left + 32;

  if ptInRect(checkRect, Point(x,y)) then
    begin
    offset := tvwFileGroups.ClientToScreen(Point(0,0));
    if tvwFileGroups.Selected.HasChildren then
      tvwFileGroups.PopupMenu := popInstallGroupSelection
    else
      tvwFileGroups.PopupMenu := popInstallSelection;

    tvwFileGroups.PopupMenu.Popup(tvwFileGroups.Selected.DisplayRect(true).Left + offset.x - 34,
                              tvwFileGroups.Selected.DisplayRect(true).Top + offset.y);
    end;
end;

procedure TfrmInstallationWizard.installToClick(Sender: TObject);
begin
  if tvwFileGroups.Selected.Data <> nil then
    begin
    if (sender = installToPC) or (sender = noInstall) or (sender = installToPalm) then
      TNodeInfo(tvwFileGroups.Selected.Data).cardIndex := -1
    else
      TNodeInfo(tvwFileGroups.Selected.Data).cardIndex := (Sender as TMenuItem).MenuIndex - 3;
    end;
  tvwFileGroups.Selected.ImageIndex := (sender as TMenuItem).ImageIndex - 4;
  tvwFileGroups.Invalidate;
  Repaint;
end;

procedure TfrmInstallationWizard.groupInstallToClick(Sender: TObject);
var
  childNode: TTreeNode;
begin
  tvwFileGroups.Selected.ImageIndex := (sender as TMenuItem).ImageIndex - 4;
  if tvwFileGroups.Selected.Data = nil then
    begin
      childNode := tvwFileGroups.Selected.getFirstChild;
      while childNode <> nil do
        begin
        if (sender as TMenuItem).ImageIndex - 4 = 3 then
          begin
          childNode.ImageIndex := 3;
          end
        else if (sender as TMenuItem).ImageIndex - 4 = 0 then
          begin
          if (childNode.ImageIndex = 3) and (childNode.Data <> nil) then
            begin
            childNode.ImageIndex := 1;
            if not (upperCase(installerSettings.installerFiles[TNodeInfo(childNode.Data).fileIndex].destination) = '$PALM_DEVICE') then
              begin
              childNode.ImageIndex := 0;
              end;
            end;
          end
        else
          begin
          if childNode.Data <> nil then
            begin
            TNodeInfo(childNode.Data).cardIndex := -1;
            end; //if childNode.Data <> nil then
          end;
        childNode := tvwFileGroups.Selected.GetNextChild(childNode);
        end; //while childNode <> nil do
      if tvwFileGroups.Selected.ImageIndex < 3 then
        tvwFileGroups.Selected.ImageIndex :=  tvwFileGroups.Selected.ImageIndex + 8;
    end; //if tvwFileGroups.Selected.Data = nil
  tvwFileGroups.Invalidate;
  Repaint;
end;

procedure TfrmInstallationWizard.cbxCurrentUserChange(Sender: TObject);
var
  ii: Integer;
  cardList: TStringList;
  miNew: TMenuItem;
  userHasCard: boolean;
  pFile: TPalmInstallerFile;
  childNode: TTreeNode;
begin
  while popInstallSelection.Items[3].Caption <> '-' do
    begin
    popInstallSelection.Items.Delete(3);
    end;
  while popInstallGroupSelection.Items[3].Caption <> '-' do
    begin
    popInstallGroupSelection.Items.Delete(3);
    end;    // while
  //set the PopMenu up
  cardList := TStringList.Create;
  installerSettings.fillCardList(installerSettings.installUsers[cbxCurrentUser.ItemIndex].index, cardList);
  userHasCard := cardList.Count > 0;
  for ii := 0 to cardList.Count - 1 do
    begin
    miNew := TMenuItem.Create(popInstallSelection);
    miNew.Caption := cardList[ii];
    miNew.OnClick := installToClick;
    miNew.ImageIndex := 6;
    popInstallSelection.Items.Insert(3, miNew);
    end; // for
  for ii := 0 to cardList.Count - 1 do
    begin
    miNew := TMenuItem.Create(popInstallGroupSelection);
    miNew.Caption := 'Install all to ' + cardList[ii];
    miNew.OnClick := groupInstallToClick;
    miNew.ImageIndex := 6;
    popInstallGroupSelection.Items.Insert(3, miNew);
    end; // for
  for ii := 0 to tvwFileGroups.Items.Count - 1 do
    begin
    if tvwFileGroups.Items[ii].Data <> nil then
      begin
      pFile := installerSettings.installerFiles[TNodeInfo(tvwFileGroups.Items[ii].Data).fileIndex];
      if not pFile.installByDefault then
        tvwFileGroups.Items[ii].ImageIndex := 3
      else if (upperCase(pFile.destination) = '$PALM_DEVICE') and (not pFile.forceCardInstallation) then
        tvwFileGroups.Items[ii].ImageIndex := 1
      else if (upperCase(pFile.destination) = '$PALM_DEVICE') and (pFile.forceCardInstallation) then
        begin
        tvwFileGroups.Items[ii].ImageIndex := 2;
        if not userHasCard then tvwFileGroups.Items[ii].ImageIndex := 3;
        end
      else
        tvwFileGroups.Items[ii].ImageIndex := 0;
      end;
    end;
  for ii := 0 to tvwFileGroups.Items.Count - 1 do
    begin
    if (tvwFilegroups.Items[ii].Level = 0) or (not tvwFileGroups.Items[ii].HasChildren) then
      continue;
    childNode := tvwFileGroups.Items[ii].getFirstChild;
    tvwFileGroups.Items[ii].ImageIndex := childNode.ImageIndex;
    while childNode <> nil do
      begin
      if tvwFileGroups.Items[ii].ImageIndex <> childNode.ImageIndex then
        if (not(tvwFileGroups.Items[ii].ImageIndex in [1,2])) and (not (childNode.ImageIndex in [1,2])) then
          begin
          tvwFileGroups.Items[ii].ImageIndex := 0;
          break;
          end
        else
          begin
          if (tvwFileGroups.Items[ii].ImageIndex = 2) and (childNode.ImageIndex = 1) then
            tvwFileGroups.Items[ii].ImageIndex := 1;
          end;
      childNode := tvwFileGroups.Items[ii].GetNextChild(childNode);
      end;
    end;

end;

procedure TfrmInstallationWizard.btnChangeDirectoryClick(Sender: TObject);
begin
  pnlDirectoryDetails.Visible := true;
  edtInstallationDirectory.Text := installerSettings.installationPath;
end;

procedure TfrmInstallationWizard.btnLookupClick(Sender: TObject);
var
  dir: string;
begin
  dir := getFirstExistingRoot(installerSettings.installationPath);
  if selectDirectory('Select the Directory for your installation project', '', dir) then
  edtInstallationDirectory.Text := IncludeTrailingPathDelimiter(dir) + installerSettings.productName;

end;

procedure TfrmInstallationWizard.edtInstallationDirectoryChange(
  Sender: TObject);
begin
  lblInstallationDirectory.Caption := edtInstallationDirectory.Text;
end;

procedure TfrmInstallationWizard.btnBackClick(Sender: TObject);
begin
  if installerSettings = nil then
    raise Exception.Create('Installation Settings have not been loaded');
  jumpToPage(getPreviousPage);
end;

procedure TfrmInstallationWizard.FormShow(Sender: TObject);
begin
  if installerSettings.runSilent then
    begin
    btnBack.Visible := false;
    btnNext.Caption := 'OK';
    if installerSettings.preInstallationMessageOverrideSilentMode then
      begin
      currentPage := ipPreInstallMessage;
      jumpToPreInstall;
      end
    else if installerSettings.userPromptOverrideSilentMode then
      begin
      currentPage := ipUserSelection;
      jumpToUserSelection;
      end
    else if installerSettings.postInstallationMessageOverrideSilentMode then
      begin
      btnNext.Visible := false;
      btnCancel.Caption := 'OK';
      currentPage := ipPostInstallMessage;
      jumpToPostInstall;
      end
    end;
end;

procedure TfrmInstallationWizard.tvwFileGroupsExpanding(Sender: TObject;
  Node: TTreeNode; var AllowExpansion: Boolean);
begin
  if Node.Count = 1 then
    if node.getFirstChild.Count = 0 then
      allowExpansion := false;
end;

procedure TfrmInstallationWizard.lbxDirectoryNamesClick(Sender: TObject);
begin
  edtShortCutDirectory.Text := includeTrailingPathDelimiter(lbxDirectoryNames.
    Items[lbxDirectoryNames.ItemIndex]) + installerSettings.shortCutPath;
end;

end.
