unit pInstallerTypes;

{$DEFINE FULL}
{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

interface
uses
  Classes, SysUtils, hxInstallUtilities, windows, Forms, registry, msxmldom, XMLDoc,
  xmldom, XMLIntf, hXxmlUtils, shellLinks{$IFDEF DEMO}, obfuscator{$ENDIF};

type TRegValueType = (rtInteger, rtString);

type ESettingsError = class(Exception);

type TFileGroup = class(TObject)
  public
    description: string;
    id: word;
    installByDefault: boolean;
    selectedForInstallation: boolean;
    hideContents: boolean;
    constructor Create(xmlNode: IXMLNode); overload;
end;

type TRegChanges = class(TObject)
  public
  rootKey: HKEY;
  hivePath: String;
  name: String;
  stringValue: String;
  intValue: integer;
  valueType: TRegValueType;
  constructor Create(xmlNode: IXMLNode); overload;
end;

type TJavaConduitSettings = class(TObject)
  public
//   jSyncInstallerFileName: String;
//   runInstallerSilently: boolean;
   ClassPath: String;
   ClassName: String;
//   JarFile: String;
   constructor Create(xmlNode: IXMLNode); overload;
   constructor Create(jSyncInstallerFileName: string; runInstallerSilently: boolean; ClassPath: string;
                      ClassName: String; JarFile: String) ; reintroduce; overload;
end;

type TComConduitSettings = class(TObject)
  public
//   comInstallerFileName: String;
//   runInstallerSilently: boolean;
   COMClient: String;
//   COMFile: String;
   constructor Create(xmlNode: IXMLNode); overload;
   constructor Create(comInstallerFileName: string; runInstallerSilently: boolean; COMClient: string) ; reintroduce; overload;

end;

type TConduitSettings = class(TObject)
  public
    fileName: String;
    creatorID: String;
    cndPriority: word;
    cndfile: String;
    cndDirectory: String;
    cndRemote: String;
    cndTitle: String;
    cndUser: String;
    cndInfo: String;
    comSettings: TComConduitSettings;
    javaSettings: TJavaConduitSettings;
    constructor Create(xmlNode: IXMLNode); overload;
end;

type TShortcutSettings = class(TObject)
  public
    fileName : string;
    description: string;
    shortCutPath: string;
    parameters : string;
    startupDirectory: string;
    shortCutKey: byte;
    shortCutKeyModifiers: byte;
    windowState: TWindowState; //showCMD
    iconPath: string;
    iconIndex: integer;
    constructor Create(fName: string; xmlNode: IXMLNode); overload;
    procedure addShortcutToSystem;
end;

type TPalmInstallerFile = class(TObject)
  public
    fileName             : string;
    executeAfterInstall  : boolean;
    executionParameters  : string;
    doNotInstall         : boolean;
    conduitSettings      : TConduitSettings;
    requiredFile         : Boolean;
    cardInstallable      : Boolean;
    forceCardInstallation: Boolean;
    registerAsNotifier   : Boolean;
    registerDll          : boolean;
    fileGroup            : Word;
    fileDescription      : string;
    destination          : string;
    installByDefault     : boolean;
    shortCut             : TShortcutSettings;
    constructor Create(xmlNode: IXMLNode; addSize: boolean = true); overload;
    destructor destroy; override;
  end;
PPalmInstallerFile = ^TPalmInstallerFile;

type THotSyncUser = class(TObject)
 public
   index: integer;
   id: integer;
   displayName: string;
   constructor create(index, id: integer; displayName: string); overload;
end;

type TInstallerUser = class(THotSyncUser)
 public
   installFileMatrix: array of SmallInt;
   installCardFileMatrix: array of SmallInt;
   function fileSelectedForInstall(fileIndex: integer): boolean;
   function fileSelectedForCardInstall(fileIndex: integer): boolean;
   procedure initializeFileMatrixes(files: array of TPalmInstallerFile);
end;

type TPalmInstaller = class(TObject)
  private
   hCndDll: HModule;
   hInstDll: HModule;
   hUserDll: HModule;
   umLoaded: Boolean;
   nmRegisterLoaded: Boolean;
   getUserDirectory: FUNCTION (userID: LongInt; pUserDir: PChar; var piSize: SmallInt): LongInt  stdcall ;
   RemoveConduit: FUNCTION(pCreatorID: PChar):LongInt stdcall;
   GetLibVersion: FUNCTION(): LongInt stdcall;
   GetExecPath: FUNCTION(pPath: PChar; piSize:Pointer):LongInt stdcall;
   InstallCreator: FUNCTION(pCreator: PChar; iType:integer):LongInt stdcall;
   SetCreatorName: FUNCTION(pCreatorID: PChar; pConduitName: PChar): LongInt stdcall;
   SetCreatorPriority: FUNCTION(pCreatorID: PChar; priority: LongInt): LongInt stdcall;
   SetCreatorFile: FUNCTION(pCreatorID: PChar; pFile: PChar): LongInt stdcall;
   SetCreatorDirectory: FUNCTION(pCreatorID: PChar; pDirectory: PChar): LongInt stdcall;
   SetCreatorRemote: FUNCTION(pCreatorID: PChar; pRemoteDB: PChar): LongInt stdcall;
   SetCreatorUser: FUNCTION(pCreatorID: PChar; pUsername: PChar): LongInt stdcall;
   SetCreatorInfo: FUNCTION(pCreatorID: PChar; pInfo: PChar): LongInt stdcall;
   SetCreatorTitle: FUNCTION(pCreatorID: PChar; pTitle: PChar): LongInt stdcall;
   SetCreatorValueString: FUNCTION(pCreatorID: PChar; pValue: PChar; pString: PChar): LongInt stdcall;
   GetHotSyncExecPath: FUNCTION(pPath: PChar; piSize:Pointer):LongInt stdcall;
   InstallPalmFile: FUNCTION(pUser: PChar; pFileSpec: PChar): LongInt stdcall;
   GetUser: FUNCTION(iIndex: LongInt; pUserBuffer: Pchar; psUserBufSiz: Pointer): LongInt stdcall;
   IsUserProfile: FUNCTION(pUser: PChar): LongInt;
   GetSlotCount: function (userId: LongInt;  var NumSlots: word): LongInt  stdcall ;
   GetSlotDisplayName: function(userId: LongInt; slotId: LongInt; SlotDisplayName: pchar; var plSize: LongInt): LongInt  stdcall ;
   GetSlotMediaType: function (userId: LongInt; slotId: LongInt; var mediaType: longInt): LongInt stdcall;
   InstallFileToCard: function(userId: LongInt; slotId: LongInt; const FileName: PChar): LongInt stdcall;
   registerNotifier: function(const FileName: PChar): LongInt stdcall;
   findNotifier: function(const FileName: PChar): LongInt stdcall;
   unRegisterNotifier: function(const FileName: PChar): LongInt stdcall;
   FRebootNeeded: boolean;
   procedure loadHotSyncUserList;
   procedure InitFunc(var funcPtr : pointer; libHandle : HModule; funcName :string);
   function InitNonCriticalFunc(var funcPtr : pointer; libHandle : HModule; funcName :string):boolean;
   procedure initializeLog;
   procedure unregisterIfNotifier(libHandle: HModule; fileName: string);
  protected
   procedure registerConduit(settings: TConduitSettings);
  protected
   loggingEnabled : Boolean;
   showFileGroupDetails: Boolean;
   showFinalThanks : boolean;
   registryFiles: TStringList;
   registryConduits: TStringList;
   regChanges: array of TRegChanges;
   appName : string;
   runJSyncExe : string;
   hideJSync : boolean;
   ComInstallerExe : string;
   hideComInstaller : boolean;
  public
   forceReboot: boolean;
   runSilent : boolean;
   userPromptOverrideSilentMode: boolean;
   preInstallationMessageOverrideSilentMode :boolean;
   postInstallationMessageOverrideSilentMode :boolean;
   uninstallString: string;
   allowUserPathSelection: boolean;
   hotSyncPath: string;
   postInstallMsgCaption: String;
   postInstallMsgTitle: String;
   postInstallMsgFile : string;
   productName : string;
   hotSyncUsers: array of THotSyncUser;
   installUsers: array of TInstallerUser;
   preInstallMsgFile : string;
   preInstallShowLicense : boolean;
   preInstallMsgCaption: string;
   preInstallMsgTitle : string;
   preInstallSetDefault: boolean;
   preInstallDefaultAs: boolean;
   allowUserShortcutPathPrompt: boolean;
   GetUserID: function (iIndex: LongInt; var userID: LongInt): LongInt  stdcall ;
   GetUserCount : FUNCTION(): LongInt stdcall;
   GetSlotInfo: function (userId: LongInt; var slotIdList: LongInt; var numEntries: Word): LongInt  stdcall ;
   installationPath: string;
   fileGroups: array of TFileGroup;
   installerFiles: Array of TPalmInstallerFile;
   shortCutPath: string;
   unInstallErrors: string;
   procedure loadLibraryFunctions;
   procedure buildUninstaller;
   procedure restartHotSync;
   procedure applyCustomRegistryChanges;
   function allUsersHaveCards: boolean;
   function conduitSelectedForInstallation: boolean;
   function pcFilesSelectedForInstallation: boolean;
   procedure runPostInstallExes;
   function allUsersHaveMatchingCard(cardName: string):boolean;
   function userHasMatchingCard(userID: integer; cardName: string): Boolean;
   procedure fillCardList(userIndex: LongInt; cardList: TStringList);
   function getFileGroupCount(groupIndex: word): word;
   function fileIsSelectedForInstallation(fileIndex: integer): boolean;
   function fileGroupIsOnlyPalmFiles(groupIndex: word): Boolean;
   function fileGroupIsCardInstallable(groupIndex: word): Boolean;
   function getSlotName(userID:LongInt; slotID: LongInt): string;
   function getInstallableSlotCount(userID: integer) : word;
   function getUserNameString(userNum: integer): String;
   procedure installFile(fileIndex: integer; userIndex: integer);
   function isInstalledUser(userID: integer): boolean;
   procedure AddLogEntry(logEntry: String);
   constructor Create; overload;
   destructor Destroy; override;
   procedure uninstall;
   procedure loadXML(xmlFileName: String);
   function execute: boolean;
   procedure initializeLibraries;
   function getHotSyncUserCount: integer;
   function palmFilesToInstall: boolean;
   function notifiersToRegister: boolean;
end;

const
CONDUIT_APPLICATION=1;
ERR_CONDUIT_MGR=-1000;
ERR_CREATORID_ALREADY_IN_USE=ERR_CONDUIT_MGR - 5;
PHS_QUIET_WM_CLOSE=1;
WM_POKEHOTSYNC=48837;
REG_PATH_ENTRY=30060;
ExpMediaType_PlugNPlay = 1886285939;
ERR_NOTIFIER_NOT_FOUND = -1021;
ERR_INVALID_POINTER = -1013;

implementation

uses
  installationWizardForm;


constructor TFileGroup.Create(xmlNode: IXMLNode);
begin
  inherited create;
  try
    selectedForInstallation := false;
    id := xmlNode.ChildNodes['id'].NodeValue;
    description := xmlNode.ChildNodes['description'].NodeValue;
    installByDefault := xmlGetChildBoolDef(xmlNode, 'installByDefault', true);
    hideContents := xmlGetChildBoolDef(xmlNode, 'hideContents', false);
  except
    on E: Exception do
      raise Exception.Create('Error loading File Groups from XML Node: ' + e.Message);
  end;
end;

constructor THotSyncUser.create(index, id: integer; displayName: string);
begin
  inherited create;
  self.id := id;
  self.index := index;
  self.displayName := displayName;
end;

constructor TJavaConduitSettings.Create(xmlNode: IXMLNode);
begin
  inherited create;
  try
    ClassPath := xmlNode.ChildNodes['ClassPath'].Text;
    ClassName := xmlNode.ChildNodes['ClassName'].Text;
  except
    on E: Exception do
      raise Exception.Create('Error loading Java settings from XML Node: ' + e.Message);
  end;
end;

constructor TJavaConduitSettings.Create(jSyncInstallerFileName: string; runInstallerSilently: boolean; ClassPath: string;
                      ClassName: String; JarFile: String);
begin
  inherited create;
  self.ClassPath := ClassPath;
  self.ClassName := ClassName;
end;

constructor TCOMConduitSettings.Create(xmlNode: IXMLNode);
begin
  inherited create;
  try
    COMClient := xmlNode.ChildNodes['COMClient'].Text;
  except
    on E: Exception do
      raise Exception.Create('Error loading COM settings from XML Node: ' + e.Message);
  end;
end;

constructor TCOMConduitSettings.Create(comInstallerFileName: string; runInstallerSilently: boolean; COMClient: string);
begin
  inherited create;
  self.COMClient := COMClient;
end;

constructor TRegChanges.Create(xmlNode: IXMLNode);
begin
  try
    if xmlNode.ChildNodes['rootKey'].Text = 'HKEY_CLASSES_ROOT' then
      rootKey := HKEY_CLASSES_ROOT
    else if xmlNode.ChildNodes['rootKey'].Text = 'HKEY_USERS' then
      rootKey := HKEY_USERS
    else if xmlNode.ChildNodes['rootKey'].Text = 'HKEY_LOCAL_MACHINE' then
      rootKey := HKEY_LOCAL_MACHINE
    else
      rootKey := HKEY_CURRENT_USER;
    hivePath := xmlNode.ChildNodes['hivePath'].Text;
    name := xmlNode.ChildNodes['Name'].Text;
    if xmlNode.ChildNodes['valueType'].Text = 'stringValue' then
      begin
      valueType := rtString;
      stringValue := xmlNode.ChildNodes['stringValue'].Text;
      end
    else
      begin
      valueType := rtInteger;
      intValue := xmlNode.ChildNodes['stringValue'].NodeValue;
      end;
  except
    on E: Exception do
      raise Exception.Create('Error loading registry settings from XML Node: ' + e.Message);
  end;
end;


constructor TShortCutSettings.Create(fName: string; xmlNode: IXMLNode);
var
  startState: string;
  modString: string;
begin
  inherited create;
  shortCutKey := 0;
  shortCutKeyModifiers := 0;
  description := xmlGetChildTextDef(xmlNode, 'description', fileName);

  parameters := xmlGetChildTextDef(xmlNode, 'parameters', '');
  startupDirectory := xmlGetChildTextDef(xmlNode, 'startupDirectory', '');
  if xmlNodeHasChildNode(xmlNode, 'shortCutKey') then
    shortCutKey := vkKeyScan(Char(xmlNode.ChildNodes['shortCutKey'].Text[1]));
  if xmlNodeHasChildNode(xmlNode, 'shortCutKeyModifiers') then
    begin
    modString := xmlNode.ChildNodes['shortCutKeyModifiers'].Text;
    if modString = 'CTRL' then
      shortCutKeyModifiers := VK_CONTROL
    else if modString = 'SHIFT' then
      shortCutKeyModifiers := VK_SHIFT
    else if modString = 'ALT' then
      shortCutKeyModifiers := VK_MENU
    else if modString = 'ALTSHIFT' then
      shortCutKeyModifiers := VK_MENU or VK_SHIFT
    else if modString = 'CTRLSHIFT' then
      shortCutKeyModifiers := VK_CONTROL or VK_SHIFT
    else if modString = 'CTRLALT' then
      shortCutKeyModifiers := VK_MENU or VK_CONTROL
    else if modString = 'CTRLALTSHIFT' then
      shortCutKeyModifiers := VK_MENU or VK_CONTROL or VK_SHIFT;
    end; //if xmlNodeHasChildNode(xmlNode, 'shortCutKeyModifiers') then
  startState := xmlGetChildTextDef(xmlNode, 'windowState', '');
  if lowerCase(startState) = 'minimized' then
    windowState := wsMinimized
  else if lowerCase(startState) = 'maximized' then
    windowState := wsMaximized
  else
    windowState := wsNormal;
  iconPath := xmlGetChildTextDef(xmlNode, 'iconPath', '');
  iconIndex := xmlGetChildIntDef(xmlNode, 'iconIndex', 0);

end;

procedure TShortCutSettings.addShortcutToSystem;
var
  lnk: TShellLink;
  fileCmd: String;
//  path: String;
begin
  if not directoryExists(shortCutPath) then
    begin
    ForceDirectories(shortCutPath);
    end;
  lnk := TShellLink.Create(IncludeTrailingPathDelimiter(shortCutPath) + description + '.lnk');
  lnk.setDescription(PChar(Description));
  fileCmd := fileName;
  if trim(parameters) <> '' then
    fileCmd := fileName + ' ' + parameters;
  lnk.setPath(fileName);
  lnk.setHotKey(shortCutKey, shortCutKeyModifiers);
  lnk.setShowCmd(Integer(windowState));
  if startupDirectory <> '' then
    lnk.setWorkingDirectory(startupDirectory);
  if iconPath <> '' then
    lnk.setIconPath(iconPath);
  if iconIndex <> 0  then
    lnk.setIconIndex(iconIndex);
  lnk.Save;
  lnk.free;

end;

constructor TConduitSettings.Create(xmlNode: IXMLNode);
begin
  inherited create;
  try
    fileName := xmlNode.ChildNodes['fileName'].Text;
    creatorID := xmlNode.ChildNodes['creatorID'].Text;
    cndPriority := xmlNode.ChildNodes['priority'].NodeValue;
    cndDirectory := xmlGetChildTextDef(xmlNode, 'directory', '');
    cndTitle := xmlGetChildTextDef(xmlNode, 'name', '');
    cndFile := xmlGetChildTextDef(xmlNode, 'file', '');
    cndRemote := xmlGetChildTextDef(xmlNode, 'remote', '');
    cndUser := xmlGetChildTextDef(xmlNode, 'user', '');
    cndInfo := xmlGetChildTextDef(xmlNode, 'info', '');
    if xmlNodeHasChildNode(xmlNode, 'COMSettings') then
      comSettings := TCOMConduitSettings.Create(xmlNode.ChildNodes['COMSettings']);
    if xmlNodeHasChildNode(xmlNode, 'JavaSettings') then
      javaSettings := TJavaConduitSettings.Create(xmlNode.ChildNodes['JavaSettings']);
  except
    on E: Exception do
      raise Exception.Create('Error loading Conduit settings from XML Node: ' + E.Message);
  end;
end;

constructor TPalmInstallerFile.Create(xmlNode: IXMLNode; addSize: boolean = true);
var
  fSize: string;
begin
  inherited create;
  try
    executeAfterInstall := false;
    fileName := xmlNode.ChildNodes['fileName'].Text;
    destination := xmlGetChildTextDef(xmlNode, 'destination', fileName);
    if xmlNodeHasChildNode(xmlNode, 'conduit') then
      conduitSettings := TConduitSettings.Create(xmlNode.ChildNodes['conduit']);
    if (destination = '$PALMUSERDIR') and (conduitSettings <> nil) then
      begin
      raise Exception.Create('Conduit Files cannot be installed to $PALMUSERDIR');
      end;
    cardInstallable := xmlGetChildBoolDef(xmlNode, 'cardInstallable', false);
    forceCardInstallation := xmlGetChildBoolDef(xmlNode, 'forceCardInstallation', false);
    registerAsNotifier := xmlGetChildBoolDef(xmlNode, 'registerAsNotifier', false);
    fileGroup := xmlGetChildIntDef(xmlNode, 'fileGroup', 0);
    requiredFile := fileGroup = 0;
    registerDll := xmlGetChildBoolDef(xmlNode, 'registerDll', false);
    fileDescription := xmlGetChildTextDef(xmlNode, 'fileDescription', fileName);
    fSize := getPrintableFileSize(fileName);
    if (fSize <> '') and addSize then
      fileDescription := fileDescription + ' (' + fSize + ')';
    installByDefault := xmlGetChildBoolDef(xmlNode, 'installByDefault', true);
    if xmlNodeHasChildNode(xmlNode, 'shortCut') then
      shortCut := TShortcutSettings.Create(fileName, xmlNode.ChildNodes['shortCut']);
    if xmlNodeHasChildNode(xmlNode, 'executeAfterInstallation') then
      begin
      executeAfterInstall := true;
      executionParameters := xmlGetChildTextDef(xmlNode.ChildNodes['executeAfterInstallation'], 'parameters', '');
      doNotInstall := xmlGetChildBoolDef(xmlNode.ChildNodes['executeAfterInstallation'], 'doNotInstall', false);
      end;
  except
    on E: Exception do
      raise Exception.Create('Error loading installation file from XML Node: ' + E.Message);
  end;
end;

destructor TPalmInstallerFile.destroy;
begin
  shortCut.free;
  inherited;
end;

procedure TInstallerUser.initializeFileMatrixes(files: array of TPalmInstallerFile);
var
  ii: Integer;
begin
  setLength(installCardFileMatrix, length(files));
  setLength(installFileMatrix, length(files));
  for ii := Low(files) to High(files) do
    begin
    installFileMatrix[ii] := -1;
    installCardFileMatrix[ii] := -1;
    if files[ii].fileGroup = 0 then
      begin
      installFileMatrix[ii] := 1;
      end;
    end; // for

end;

function TInstallerUser.fileSelectedForInstall(fileIndex: integer): boolean;
begin
  try
    result := installFileMatrix[fileIndex] <> -1;
  except
    result := false;
  end;
end;

function TInstallerUser.fileSelectedForCardInstall(fileIndex: integer): boolean;
begin
  try
    result := installCardFileMatrix[fileIndex] <> -1;
  except
    result := false;
  end;
end;

procedure TPalmInstaller.loadHotSyncUserList;
var
  userIndex: integer;
  id: LongInt;
begin
  for userIndex := 0 to getUserCount - 1 do
    begin
    getUserID(userIndex, id);
    SetLength(HotSyncUsers, Length(HotSyncUsers) + 1);
    hotSyncUsers[Length(HotSyncUsers) - 1] := THotSyncUser.create(userIndex, id, getUserNameString(userIndex));
    end;
end;    //

constructor TPalmInstaller.Create;
begin
  inherited;
  FRebootNeeded := false;
  runSilent := false;
  allowUserPathSelection := false;
  userPromptOverrideSilentMode := false;
  showFinalThanks := true;
  registryFiles := TStringList.Create;
  registryConduits := TStringList.Create;
  uninstallString := '<?xml version="1.0" encoding="UTF-8"?>' +
                     '<pUninstaller xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" ' +
                     'xsi:noNamespaceSchemaLocation="http://www.handx.net/devtool/pInstaller/' +
                     'schema/pUnInstaller.xsd">';
end;

destructor TPalmInstaller.Destroy;
var
  fileIndex: integer;
begin
  for fileIndex := low(installerFiles) to high(installerFiles) do
    installerFiles[fileIndex].Free;
  if hCndDll > HINSTANCE_ERROR then
    FreeLibrary(hCndDll);
  if hInstDll > HINSTANCE_ERROR then
    FreeLibrary(hInstDll);
  if hUserDll > HINSTANCE_ERROR then
    FreeLibrary(hUserDll);
  registryFiles.Free;
  registryConduits.Free;
  inherited;
end;


{------------------------------------------------------------------------------------------
*******************************************************************************************
procedure TPalmInstaller.uninstall
Description:
Handles the uninstall process for pInstaller-
Removes the conduit from HotSync
Deletes all installed PC files
Removes registry key

Needed: At the end of the Uninstall, HotSync should be restarted to remove the conduit
from the list, but adding this code causes the UnInstall to hang in Win2K
*******************************************************************************************
------------------------------------------------------------------------------------------}
procedure TPalmInstaller.uninstall;
var
  sr: TSearchRec;
  rebootRequired : boolean;
  xmlDocument: TXMLDocument;
  reg: TRegistry;
  ii: integer;
  hDll: HModule;
  att: string;
  unRegisterDll: function: HResult stdcall;
  conduitRemoved: boolean;
//  hotSyncPath: string;
  reboot: boolean;
begin
  reboot := false;
  if not fileExists(paramStr(2)) then
    begin
    Application.MessageBox('Unable to locate uninstaller data', 'Error', MB_ICONERROR);
    exit;
    end;
  unInstallErrors := '';
  conduitRemoved := false;
  hotSyncPath := '';
  xmlDocument := TXMLDocument.Create(Application);
  try
    xmlDocument.FileName := paramStr(2);
    xmlDocument.Active := true;
    with xmlDocument.DocumentElement do
      begin
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'hotsyncpath' then
          begin
          hotSyncPath := ChildNodes[ii].Text;
          end;
        end; //for ii := 0 to ChildNodes.Count - 1 do
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'unnotify' then
          begin
          unregisterIfNotifier(hCndDll, ChildNodes[ii].Attributes['notifierPath']);
          end; //if lowercase(ChildNodes[ii].NodeName) = 'unnotify' then
        end; //for ii := 0 to ChildNodes.Count - 1 do
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'removeconduit' then
          if DirectoryExists(hotSyncPath) and FileExists(IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll') then
            begin
            hCndDll := LoadLibraryEx(PChar(IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll'),0,0);
            If hCndDll<=HINSTANCE_ERROR then raise EConduitRegistryError.Create('Problem calling Conduit Manager library');
            //take care of the 4.0.1 "feature" (see note in main program code)
            try
              @GetLibVersion := GetProcAddress(hCndDll, 'CmGetLibVersion');
              SetLastError(0);
              @RemoveConduit :=GetProcAddress(hCndDll, 'CmRemoveConduitByCreatorID');
              //new extra paranoid error checking courtesy of Steve Belkin
              if (GetLastError <> 0) or (not assigned(RemoveConduit)) then
                unInstallErrors := unInstallErrors + 'Problem calling Conduit Manager library from '
                  + IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll.  Conduit not unregistered';
              conduitRemoved := true;
              att := ChildNodes[ii].Attributes['creatorID'];
              RemoveConduit(PChar(att));
            finally
              FreeLibrary(hCndDll);
            end;
          end //if DirectoryExists(hotSyncPath) and FileExists(IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll') then
        else
          begin
          unInstallErrors := unInstallErrors + 'Unable to load ' +
                            IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll' + #10#13;
          end;
        end; //for ii := 0 to ChildNodes.Count - 1 do
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'unregisterdll' then
          begin
          att := ChildNodes[ii].Attributes['dllPath'];
          hDll := LoadLibrary(PChar(att));
          try
            @unRegisterDll := GetProcAddress(hDll, 'DllUnregisterServer');
          if unRegisterDll <> 0 then
            begin
            unInstallErrors := 'Error unRegistering dll : ' + att;
            end;
          finally
            FreeLibrary(hDll);
          end;
          end; //if lowercase(ChildNodes[ii].NodeName) = 'unnotify' then
        end; //for ii := 0 to ChildNodes.Count - 1 do
      //delete files and registry entries
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'deletion' then
          begin
          if lowerCase(Variant(ChildNodes[ii].Attributes['deletionType'])) = 'file' then
            begin
            att := ChildNodes[ii].Attributes['deletionPath'];
            OutPutDebugString(PChar(att));
            rebootRequired := KillFile(ChildNodes[ii].Attributes['deletionPath']) = ioRebootRequired;
            reBoot := reBoot or rebootRequired;
            end
          else if lowerCase(ChildNodes[ii].Attributes['deletionType']) = 'registrykey' then
            begin
            reg := tregistry.create(key_read or key_write);
            try
            reg.RootKey := HKEY_LOCAL_MACHINE;
            if (ChildNodes[ii].Attributes['deletionPath'] <> '') and
                 (reg.KeyExists(ChildNodes[ii].Attributes['deletionPath'])) then
              reg.DeleteKey(ChildNodes[ii].Attributes['deletionPath']);
            finally
              freeAndNil(reg);
            end;
            end
          else if lowerCase(ChildNodes[ii].Attributes['deletionType']) = 'registryvalue' then
            begin
            reg := tregistry.create(key_read or key_write);
            try
            reg.RootKey := HKEY_LOCAL_MACHINE;
            if (reg.ValueExists(ChildNodes[ii].Attributes['deletionPath'])) and
                        (ChildNodes[ii].Attributes['deletionPath'] <> '') then
              reg.DeleteValue(ChildNodes[ii].Attributes['deletionPath']);
            finally
              freeAndNil(reg);
            end;
            end

          end; //if lowercase(ChildNodes[ii].NodeName) = 'unnotify' then
        end; //for ii := 0 to ChildNodes.Count - 1 do
      //delete directories
      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if lowercase(ChildNodes[ii].NodeName) = 'deletion' then
          begin
          if lowerCase(ChildNodes[ii].Attributes['deletionType']) = 'directory' then
            begin
            if (FindFirst(IncludeTrailingPathDelimiter(ChildNodes[ii].Attributes['deletionPath']) + '*.*', faAnyFile, sr) = 0) then
              begin
              SysUtils.FindClose(sr);
              KillDirectory(ChildNodes[ii].Attributes['deletionPath']);
              end
            else //directory is not empty
              begin
              SysUtils.FindClose(sr);
              unInstallErrors := unInstallErrors + 'Directory ' +
                      ChildNodes[ii].Attributes['deletionPath'] +
                      ' is not empty.' + #10#13;
              end;
            end
          end; //if lowercase(ChildNodes[ii].NodeName) = 'unnotify' then
        end; //for ii := 0 to ChildNodes.Count - 1 do
      end; //with xmlDocument.DocumentElement do
  finally
    freeAndNil(xmlDocument);
  end;
  decreaseUsageCount(paramStr(0), rebootRequired);
  reboot := reboot or rebootRequired;
  KillFile(paramStr(2));
  if reboot then
    begin
    if Application.MessageBox('Uninstall successful.  Your computer needs ' +
                           'to be rebooted for the files to be removed.  Do you want to restart now?',
                           'Reboot Needed', MB_YESNO or MB_ICONQUESTION) = idYes
                           then
      begin
      NTSetPrivilege('SeShutdownPrivilege', true);
      ExitWindowsEx(EWX_REBOOT + EWX_FORCE,0);
      end;
    end
  else
    begin
    if conduitRemoved then
      begin
      Application.MessageBox('Please restart HotSync to complete the removal of this product',
          'pInstaller removal', mb_iconinformation or mb_ok);
      end
    else
      Application.MessageBox('Uninstall Successful.', 'Complete', MB_OK);
    end;

end;

function TPalmInstaller.allUsersHaveCards: boolean;
var
  ii: integer;
begin
  result := true;
  for ii := Low(installUsers) to High(InstallUsers) do
    begin
    if (getInstallableSlotCount(installUsers[ii].id) <1) then
      begin
      result := false;
      exit;
      end;
    end; // for
end;

function TPalmInstaller.pcFilesSelectedForInstallation: boolean;
var
  fileIndex, userIndex: integer;
begin
  result := false;
  for fileIndex := low(installerFiles) to high(installerFiles) do
    begin
    if installerFiles[fileIndex].destination = '$PALM_DEVICE' then
      continue;
    try
      if installerFiles[fileIndex].fileGroup = 0 then
        begin
        result := true;
        exit;
        end;
      for userIndex := Low(installUsers) to High(installUsers) do
        begin
        if installUsers[userIndex].installFileMatrix[fileIndex] >= 1 then
          begin
          result := true;
          exit;
          end;
        end; // for
    except
      result := false;
    end;
    end;
end;

function TPalmInstaller.conduitSelectedForInstallation: boolean;
var
  fileIndex, userIndex: integer;
begin
  result := false;
  for fileIndex := low(installerFiles) to high(installerFiles) do
    begin
    if installerFiles[fileIndex].conduitSettings = nil then
      continue;
    try
      for userIndex := Low(installUsers) to High(installUsers) do
        begin
        if installUsers[userIndex].installFileMatrix[fileIndex] >= 1 then
          begin
          result := true;
          exit;
          end;
        end; // for
    except
      result := false;
    end;
    end;
end;

function TPalmInstaller.userHasMatchingCard(userID: integer; cardName: string): Boolean;
var
  ii: integer;
begin
  result := false;
  for ii := 0 to getInstallableSlotCount(userID) - 1 do
    begin
    if trim(lowerCase(getSlotName(userID, ii))) = trim(lowerCase(cardName)) then
      result := true;
    end; // for
end;

function TPalmInstaller.allUsersHaveMatchingCard(cardName: string):boolean;
var
  ii: integer;
  foundSlot: boolean;
begin
  result := true;
  for ii := Low(installUsers) to High(InstallUsers) do
    begin
    foundSlot := userHasMatchingCard(installUsers[ii].id, cardName);
    if (not foundSlot) and (getInstallableSlotCount(installUsers[ii].id) > 0)  then
      begin
      result := false;
      exit;
      end;
    end; // for
end;    //

procedure TPalmInstaller.unregisterIfNotifier(libHandle: HModule; fileName: string);
var
  err: longInt;
begin
  @findNotifier := GetProcAddress(hCndDll, 'NmFind');
  @unRegisterNotifier := GetProcAddress(hCndDll, 'NmUnregister');
  err := findNotifier(pchar(fileName));
  if (err <> ERR_NOTIFIER_NOT_FOUND) and (err <> ERR_INVALID_POINTER) then
    //now we know it's a notifier and we should unregister it.
    begin
    unRegisterNotifier(PChar(fileName));
    end;
end;

procedure TPalmInstaller.initializeLibraries;
var
  pathName: PChar;
  size: integer;
  cmLibName: String;
  iaLibName: String;
  umLibName: String;
begin
  try
    AddLogEntry('Initializing Libraries');
    hCndDll := LoadLibraryEx('CondMgr.dll',0,0);
    If hCndDll <= HINSTANCE_ERROR then raise EConduitRegistryError.Create('Problem calling intial library');

     //*******************************************************************************
     //there seems to be a change in the 4.0.1 desktop's condmgr.dll that causes the first
     //getprocaddress to fail, while others succeed.  This is a spectacularly lame way
     //to get around that.
     //NOTE: this is not a known problem with InstAide.dll, I'm just being paranoid
     //*******************************************************************************

    @GetLibVersion := GetProcAddress(hCndDll, 'CmGetLibVersion');
    SetLastError(0);
    @GetExecPath := GetProcAddress(hCndDll, 'CmGetHotSyncExecPath');
    if not assigned(GetExecPath) then raise EConduitRegistryError.Create('Problem calling intial library');
    pathName  := StrAlloc(MAX_PATH);
    size := MAX_PATH;
    GetExecPath(pathName, addr(size));
    hotSyncPath := extractFilePath(pathName);
    if not DirectoryExists(hotSyncPath) then
      begin
      Application.MessageBox('Unable to locate HotSync installation directory.  Check your Palm software installation.',
                          'Unable to install software', MB_ICONERROR);
                          halt;
      end;
    cmLibName:= IncludeTrailingPathDelimiter(hotSyncPath) + 'CondMgr.dll';
    iaLibName := IncludeTrailingPathDelimiter(hotSyncPath) + 'InstAide.dll';
    umLibName := IncludeTrailingPathDelimiter(hotSyncPath) + 'UserData.dll';

    hUserDll := LoadLibraryEx(PChar(umLibName),0,0);
    umLoaded := hUserDll > HINSTANCE_ERROR;
    If not umLoaded then raise EConduitRegistryError.Create('Problem calling userManager library from ' +
                                                          umLibName);

    hInstDll := LoadLibraryEx(PChar(iaLibName),0,0);
    If hInstDll <= HINSTANCE_ERROR then raise EConduitRegistryError.Create('Problem calling install aide library from ' +
                                                          iaLibName);

    if FileExists(cmLibName) then
      begin
      FreeLibrary(hCndDll);
      hCndDll := LoadLibraryEx(PChar(cmLibName),0,0);
      If hCndDll<=HINSTANCE_ERROR then raise EConduitRegistryError.Create('Problem calling install aide library from ' +
                                                          cmLibName);
      end; //if FileExists(cmLibName) then
    StrDispose(pathName);
    loadLibraryFunctions;
    loadHotSyncUserList;
  except
    on E: Exception do
      begin
      Application.MessageBox(PChar(E.Message + '.  Installation archive may be incomplete or corrupted.'),
                                  'pInstaller Error', MB_ICONERROR);
      halt;
      end;
  end;
end;

procedure TPalmInstaller.loadXML(xmlFileName: String);
var
  xmlDocument: TXMLDocument;
  InstallerFilesNode: IXMLNode;
  regNode: IXMLNode;
  ii: integer;
begin
  xmlDocument := TXMLDocument.Create(Application);
  {$IFDEF DEMO}
   xmlDocument.XML.Text := loadObfuscatedFile(xmlFileName);
  {$ELSE}
  xmlDocument.XML.LoadFromFile(xmlFileName);
  {$ENDIF}
  xmlDocument.Active := true;
  try
    with xmlDocument.DocumentElement do
      begin
      loggingEnabled := xmlGetChildBoolDef(xmlDocument.DocumentElement,
                                        'genDebugLog', false);
      if loggingEnabled then initializeLog;
      forceReboot := xmlGetChildBoolDef(xmlDocument.DocumentElement, 'forceRebootAfterInstall',
                                        false);
      productName := xmlGetChildTextDef(xmlDocument.DocumentElement, 'productName', '');
      if productName = '' then raise EConduitRegistryError.Create('Product Name must be defined.');

      //this is really only now relevant for the purposes of checking if the installation file is valid
      //it may be depricated in the future
{      installationType := stringToInstallationType(xmlGetChildTextDef(xmlDocument.DocumentElement,
                                                   'installationType', ''));
}

      showFileGroupDetails := xmlGetChildBoolDef(xmlDocument.DocumentElement, 'showFileGroupDetails',
                                                false);
      if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'preInstallationMessage') then
        begin
        preInstallMsgFile := xmlGetChildTextDef(childNodes['preInstallationMessage'],
                                                   'fileName', '');
        preInstallShowLicense := xmlGetChildBoolDef(childNodes['preInstallationMessage'],
                                                   'showLicenseAgreementRadioButtons', false);
        preInstallMsgCaption := xmlGetChildTextDef(childNodes['preInstallationMessage'],
                                                   'caption', '');
        preInstallMsgTitle := xmlGetChildTextDef(childNodes['preInstallationMessage'],
                                                   'title', productName);
        preInstallSetDefault := xmlNodeHasChildNode(childNodes['preInstallationMessage'], 'defaultAcceptanceTo');
        preInstallDefaultAs := xmlGetChildTextDef(childNodes['preInstallationMessage'], 'defaultAcceptanceTo',
                                                   'DoNotAccept') = 'Accept';
        end; //if xmlHasChildNode(xmlDocument.DocumentElement, 'preInstallationMessage') then

      if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'postInstallationMessage') then
        begin
        postInstallMsgCaption := xmlGetChildTextDef(childNodes['postInstallationMessage'],
                                                   'caption', '');
        postInstallMsgTitle := xmlGetChildTextDef(childNodes['postInstallationMessage'],
                                                   'title', productName);
        postInstallMsgFile := xmlGetChildTextDef(childNodes['postInstallationMessage'],
                                                   'fileName', '');
        end; //if xmlHasChildNode(xmlDocument.DocumentElement, 'postInstallationMessage') then
      showFinalThanks := xmlGetChildBoolDef(xmlDocument.DocumentElement,
                                            'showFinalThankYouMessage', false);
      runSilent := xmlGetChildBoolDef(xmlDocument.DocumentElement,
                                            'silentMode', false);

      installationPath := xmlGetChildTextDef(xmlDocument.DocumentElement, 'defaultPCInstallationPath', '');
      allowUserPathSelection := xmlGetChildBoolDef(xmlDocument.DocumentElement, 'allowUserPathSelection', false);

      if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'silentModeOverrides') then
        begin
        userPromptOverrideSilentMode := xmlGetChildBoolDef(xmlDocument.DocumentElement.ChildNodes['silentModeOverrides'],
                                        'showUserSelection', false);
        preInstallationMessageOverrideSilentMode := xmlGetChildBoolDef(xmlDocument.DocumentElement.ChildNodes['silentModeOverrides'],
                                        'showPreInstallationMessage', false);
        postInstallationMessageOverrideSilentMode := xmlGetChildBoolDef(xmlDocument.DocumentElement.ChildNodes['silentModeOverrides'],
                                        'showPostInstallationMessage', false);
        end; //if xmlHasChildNode(xmlDocument.DocumentElement, 'silentModeOverrides') then
      if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'shortCutDefaults') then
        begin
        shortCutPath := xmlGetChildTextDef(childNodes['shortCutSettings'], 'defaultShortCutPath',
                                        productName);
        allowUserShortcutPathPrompt := xmlGetChildBoolDef(childNodes['shortCutSettings'], 'allowUserShortcutPathPrompt',
                                        true);
        end; //if xmlHasChildNode(xmlDocument.DocumentElement, 'shortCutSettings') then

//      if not xmlNodeHasChildNode(xmlDocument.DocumentElement, 'installationType') then
//        raise EConduitRegistryError.Create('No files have been defined for installation.');

      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if ChildNodes[ii].NodeName = 'fileGroup' then
          begin
          SetLength(fileGroups, length(fileGroups) + 1);
          fileGroups[length(fileGroups)-1] := TFileGroup.Create(ChildNodes[ii]);
          end;
        end;
      InstallerFilesNode := ChildNodes['installationFile'];


      for ii := 0 to ChildNodes.Count - 1 do
        begin
        if ChildNodes[ii].NodeName = 'installationFile' then
          begin
          SetLength(installerFiles, length(installerFiles) + 1);
          installerFiles[length(installerFiles)-1] := TPalmInstallerFile.Create(ChildNodes[ii]);
          if not FileExists(installerFiles[length(installerFiles)-1].fileName) then
            begin
            raise Exception.Create('Unable to locate installation file "' +
              installerFiles[length(installerFiles)-1].fileName + '"');
            end;
          end;
        end;
      InstallerFilesNode := ChildNodes['installationFile'];

      if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'registryChanges') then
        begin
        regNode := xmlDocument.DocumentElement.ChildNodes['registryChanges'];
        //registryConduitFilename := xmlGetChildTextDef(regNode, 'conduitReplacementFilename', '');
        for ii := 0 to regNode.ChildNodes.Count - 1 do
          begin
          if regNode.ChildNodes[ii].NodeName = 'registryFile' then
              begin
              registryFiles.Add(regNode.ChildNodes[ii].ChildNodes['fileName'].Text);
              registryConduits.Add(xmlGetChildTextDef(regNode.ChildNodes[ii],
                            'conduitReplacementFilename', ''));
              end
          else
            begin
            SetLength(regChanges, Length(regChanges) + 1);
            regChanges[length(regChanges)-1] := TRegChanges.Create(regNode.ChildNodes[ii]);
            end; //if regNode.ChildNodes[ii].NodeName = 'fileName' then.. else
          end; // for ii := 0 to regNode.ChildNodes.Count - 1 do
        end; //if xmlNodeHasChildNode(xmlDocument.DocumentElement, 'registryChanges') then
        if registryFiles.Count <> registryConduits.Count then
          begin
          raise Exception.Create('Error loading Registry Settings from XML Node: Counts do not match');
          end;
      end; //with xmlDocument.DocumentElement do
  except
    on E: Exception do
      begin
      xmlDocument.Active := false;
      freeAndNil(xmlDocument);
      raise ESettingsError.Create('Error in Conduit Settings: ' + E.Message);
      end; //on E: Exception do
  end; //try.. except
  xmlDocument.Active := false;
  freeAndNil(xmlDocument);
end;

function TPalmInstaller.notifiersToRegister: boolean;
var
  ii : Integer;
begin
  result := false;
  for ii:= low(installerFiles) to high(installerFiles) do
    begin
    if installerFiles[ii].registerAsNotifier then
      begin
      result := true;
      break;
      end; //if installerFiles[ii].registerAsNotifier then
    end; //for ii:= low(installFiles) to high(installFiles) do
end; //function TPalmInstaller.notifiersToRegister: boolean;

function TPalmInstaller.palmFilesToInstall: boolean;
var
  ii : Integer;
begin
  result := false;
  for ii:= low(installerFiles) to high(installerFiles) do
    begin
    if uppercase(installerFiles[ii].destination) = '$PALM_DEVICE' then
      begin
      result := true;
      break;
      end; //if uppercase(installerFiles[ii].destination) = '$PALM_DEVICE' then
    end; //for ii:= low(installFiles) to high(installFiles) do
end; //function TPalmInstaller.palmFilesToInstall: boolean;


{------------------------------------------------------------------------------------------
*******************************************************************************************
procedure TPalmInstaller.InitFunc
Description:
Loads a library function into a pointer, raising an error if needed

Input:
funcPtr : a procedural type variable for the function to load
libHandle : an HModule that has already been initialized with a call to LoadLibraryEx
funcName: the Name of the function to be loaded

*******************************************************************************************
------------------------------------------------------------------------------------------}
procedure TPalmInstaller.InitFunc(var funcPtr : pointer; libHandle : HModule; funcName :string);
var
  funcLoadError : boolean;
begin
  //this nice, clean function courtesy of Steve Belkin
   FuncPtr:= GetProcAddress(LibHandle, pChar(funcName));
   funcLoadError := (GetLastError <> 0) or (not assigned(FuncPtr));
   if funcLoadError then raise EConduitRegistryError.Create('Problem calling '+funcName);
end;

{------------------------------------------------------------------------------------------
*******************************************************************************************
function TPalmInstaller.InitNonCriticalFunc
Description:
Loads a library function into a pointer, returns false if it doesn't load
this is the same as InitFunc, except it doesn't raise errors
It was added to load the UserData.dll functions, which are only needed for
card install support, and therefore, shouldn't result in no installation

Input:
funcPtr : a procedural type variable for the function to load
libHandle : an HModule that has already been initialized with a call to LoadLibraryEx
funcName: the Name of the function to be loaded

Return:
A boolean indicating success or failure

*******************************************************************************************
------------------------------------------------------------------------------------------}
function TPalmInstaller.InitNonCriticalFunc(var funcPtr : pointer; libHandle : HModule; funcName :string): boolean;
var
  funcLoadError : boolean;
begin
  //this nice, clean function courtesy of Steve Belkin
   FuncPtr:= GetProcAddress(LibHandle, pChar(funcName));
   funcLoadError := (GetLastError <> 0) or (not assigned(FuncPtr));
   result := not funcLoadError;
end;


procedure TPalmInstaller.initializeLog;
var
  logFile: TextFile;
begin
  AssignFile(logFile, 'c:\hXinstall.log');
  ReWrite(logFile);
  WriteLn(logFile, 'Beginning Install');
  WriteLn(logFile, GetFileVersionString(paramStr(0)));
  WriteLn(logFile, '_______________________________________________________');
  CloseFile(logFile);
end;

procedure TPalmInstaller.AddLogEntry(logEntry: String);
var
  logFile: textfile;
begin
  if not loggingEnabled then exit;
  AssignFile(logFile, 'c:\hXinstall.log');
  Append(logFile);
  WriteLn(logFile, logEntry);
  CloseFile(logFile);
end;

procedure TPalmInstaller.loadLibraryFunctions;
begin

   //*******************************************************************************
   //there seems to be a change in the 4.0.1 desktop's condmgr.dll that causes the first
   //getprocaddress to fail, while others succeed.  This is a spectacularly lame way
   //to get around that.
   //*******************************************************************************
   @GetLibVersion := GetProcAddress(hCndDll, 'CmGetLibVersion');
   SetLastError(0);

   AddLogEntry('loading dll functions');
   InitFunc(@InstallCreator, hCndDll, 'CmInstallCreator');
   InitFunc(@SetCreatorName, hCndDll, 'CmSetCreatorName');
   InitFunc(@SetCreatorPriority, hCndDll, 'CmSetCreatorPriority');
   InitFunc(@SetCreatorFile, hCndDll, 'CmSetCreatorFile');
   InitFunc(@SetCreatorDirectory, hCndDll, 'CmSetCreatorDirectory');
   InitFunc(@SetCreatorRemote, hCndDll, 'CmSetCreatorRemote');
   InitFunc(@SetCreatorValueString, hCndDll, 'CmSetCreatorValueString');
   InitFunc(@SetCreatorUser, hCndDll, 'CmSetCreatorUser');
   InitFunc(@SetCreatorInfo, hCndDll, 'CmSetCreatorInfo');
   InitFunc(@SetCreatorTitle, hCndDll, 'CmSetCreatorTitle');
   InitFunc(@GetHotSyncExecPath, hCndDll, 'CmGetHotSyncExecPath');
   nmRegisterLoaded := InitNonCriticalFunc(@registerNotifier, hCndDll, 'NmRegister');
   if (not nmRegisterLoaded) and (notifiersToRegister) then
     begin
     EConduitRegistryError.Create('Unable to register notifiers with this '+
           'version of the Palm Desktop.  Please upgrade your Palm Desktop ' +
           'before installing this product.');
     end; //if (not nmRegisterLoaded) and (notifiersToRegister) then
   AddLogEntry('loading install aide functions');
   @IsUserProfile := GetProcAddress(hInstDll, 'IsUserProfile');
   SetLastError(0);

   InitFunc(@InstallPalmFile, hInstDll, 'PltInstallFile');
   InitFunc(@GetUserCount, hInstDll, 'PltGetUserCount');
   InitFunc(@GetUser, hInstDll, 'PltGetUser');

   AddLogEntry('loading User Manager functions');
   InitFunc(@GetUserID, hUserDll, 'UmGetUserID');
   InitFunc(@GetUserDirectory, hUserDll, 'UmGetUserDirectory');
   if umLoaded then
     umLoaded := InitNonCriticalFunc(@GetSlotCount, hUserDll, 'UmSlotGetSlotCount');
   if umLoaded then
     umLoaded := InitNonCriticalFunc(@GetSlotInfo, hUserDll, 'UmSlotGetInfo');
   if umLoaded then
     umLoaded := InitNonCriticalFunc(@installFileToCard, hInstDll, 'PlmSlotInstallFile');
   if umLoaded then
     umLoaded := InitNonCriticalFunc(@GetSlotDisplayName, hUserDll, 'UmSlotGetDisplayName');
   if umLoaded then
     umLoaded := InitNonCriticalFunc(@GetSlotMediaType, hUserDll, 'UmSlotGetMediaType');

end;

function TPalmInstaller.getUserNameString(userNum: integer): String;
var
  size: Integer;
  userInstallName: PChar;
begin
  size := MAX_PATH;
  userInstallname := StrAlloc(size-1);
  size := 1;
  GetUser(userNum, userInstallName, addr(size));   //first call fails, but returns correct size
  GetUser(userNum, userInstallName, addr(size));
  result := string(userInstallName);
  strDispose(userInstallName);
end;

procedure TPalmInstaller.installFile(fileIndex: integer; userIndex: integer);
var
  pathName: PChar;
  size: SmallInt;
  userDir: string;
  basePath: string;
  path: string;
  hDll: THandle;
  registerDll: function: HResult stdcall;
  ii: integer;
  str: pchar;
begin
  AddLogEntry('Installing ' + installerFiles[fileIndex].fileName);
  if uppercase(installerFiles[fileIndex].destination) = '$PALM_DEVICE' then
    begin
    if (installUsers[userIndex].installCardFileMatrix[fileIndex] = -1) then
      begin
      AddLogEntry('Installing ' + installerFiles[fileIndex].fileName + ' to palm');
      InstallPalmFile(PChar(installUsers[userIndex].displayName),
         PChar(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0))) + installerFiles[fileIndex].fileName));
      end
    else
      begin
      AddLogEntry('Installing ' + installerFiles[fileIndex].fileName + ' to card');
      str := PChar(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0))) + installerFiles[fileIndex].fileName);
      InstallFileToCard(LongInt(InstallUsers[userIndex].id),
             word(installUsers[userIndex].installCardFileMatrix[fileIndex]),
             str);
      end;
    end
  else //if destination not palm device
    begin
    path := installerFiles[fileIndex].destination;
    path := StringReplace(path, '$USERINSTALLDIR\', IncludeTrailingPathDelimiter(installationPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$HOTSYNCDIR\', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$SYSDIR\', IncludeTrailingPathDelimiter(getSystemPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$WINDIR\', IncludeTrailingPathDelimiter(getWindowsPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$USERINSTALLDIR', IncludeTrailingPathDelimiter(installationPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$HOTSYNCDIR', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$HOTSYNCDIR', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$SYSDIR', IncludeTrailingPathDelimiter(getSystemPath), [rfReplaceAll, rfIgnoreCase]);
    path := StringReplace(path, '$WINDIR', IncludeTrailingPathDelimiter(getWindowsPath), [rfReplaceAll, rfIgnoreCase]);

    if pos('$PALMUSERDIR', path) = 0 then
      begin
      //copy file
      if not directoryExists(path) then
        begin
        AddLogEntry('Creating directory ' + path);
        forceDirectories(path);
        uninstallString := uninstallString + '<deletion deletionType="Directory" deletionPath="' +
                           path + '"/>';
        end; //if not directoryExists(path) then
      AddLogEntry('Installing ' + installerFiles[fileIndex].fileName + ' to ' + path);
      FRebootNeeded := FRebootNeeded or (hxInstallUtilities.installFile(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0)))+
                               installerFiles[fileIndex].fileName,
                               IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName) = ioRebootRequired);
      uninstallString := uninstallString + '<deletion deletionType="File" deletionPath="' +
                         IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName + '"/>';
      end
    else
      begin
      basePath := path;
      for ii := low(installUsers) to high(installUsers) do
        begin
        pathName  := StrAlloc(MAX_PATH);
        size := MAX_PATH;
        GetUserDirectory(installUsers[ii].id ,pathName, size);
        userDir := pathName;
        StrDispose(pathName);
        path := StringReplace(basePath, '$PALMUSERDIR', userDir, [rfReplaceAll, rfIgnoreCase]);

        if not directoryExists(path) then
          begin
          AddLogEntry('Creating directory ' + path);
          forceDirectories(path);
          uninstallString := uninstallString + '<deletion deletionType="Directory" deletionPath="' +
                             path + '"/>';
          end;
        AddLogEntry('Installing ' + installerFiles[fileIndex].fileName + ' to ' + path);
        FRebootNeeded := FRebootNeeded or (hxInstallUtilities.installFile(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0)))+
                                 installerFiles[fileIndex].fileName,
                                 IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName) = ioRebootRequired);

        uninstallString := uninstallString + '<deletion deletionType="File" deletionPath="' +
                           IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName + '"/>';

        end; //for ii := low(installUsers) to high(installUsers) do
      end; //if not directoryExists(path) then.. else
    //register conduit
    if installerFiles[fileIndex].conduitSettings <> nil then
      begin
      installerFiles[fileIndex].conduitSettings.fileName := IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName;
      AddLogEntry('Registering ' + installerFiles[fileIndex].fileName + ' as conduit');
      registerConduit(installerFiles[fileIndex].conduitSettings);
      uninstallString := uninstallString + '<removeConduit conduitPath="' +
                       IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName + '" creatorID="' +
                       installerFiles[fileIndex].conduitSettings.creatorID + '"/>';

      end;
    //register dll
    if installerFiles[fileIndex].registerDll then
      begin
      AddLogEntry('Registering ' + installerFiles[fileIndex].fileName + ' with Windows');
      hDll := LoadLibrary(PChar(IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName));
      if hDll = 0 then application.MessageBox('Error registering dll', 'pInstaller', mb_ok);
      try
        @registerDll := GetProcAddress(hDll, 'DllRegisterServer');
        if @registerDll = Nil then
          application.MessageBox('Error registering dll', 'pInstaller', mb_ok);
      if registerDll = 0 then
        begin
        uninstallString := uninstallString + '<unRegisterDll dllPath="' +
                       IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName + '"/>';
        end
      else
        begin
        application.MessageBox('Error registering dll', 'pInstaller', mb_ok);
        end;
        //raise Exception.CreateFmt(SRegFail, [ProcName[RegAction], FileName]);
      finally
        FreeLibrary(hDll);
      end;
      end;

    //register notifier
    if installerFiles[fileIndex].registerAsNotifier and nmRegisterLoaded then
      begin
      AddLogEntry('Registering ' + installerFiles[fileIndex].fileName + ' as notifier');
      registerNotifier(PChar(IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName));
      uninstallString := uninstallString + '<unNotify deletionPath="' +
                       IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName + '"/>';
      end; //if installerFiles[fileIndex].registerAsNotifier and nmRegisterLoaded then

    //create shortcut
    if installerFiles[fileIndex].shortCut <> nil then
      begin
      AddLogEntry('Creating shortcut for ' + installerFiles[fileIndex].fileName);
      if not directoryExists(IncludeTrailingPathDelimiter(getStartProgramsMenuPath) + shortCutPath) then
        begin
        uninstallString := uninstallString + '<deletion deletionType="Directory" deletionPath="' +
                           getFirstNonExistingDirectory(IncludeTrailingPathDelimiter(getStartProgramsMenuPath) + shortCutPath) + '"/>';
        end;
      installerFiles[fileIndex].shortCut.fileName := IncludeTrailingPathDelimiter(path)+installerFiles[fileIndex].fileName;
      installerFiles[fileIndex].shortCut.shortCutPath := IncludeTrailingPathDelimiter(getStartProgramsMenuPath) + shortCutPath;
      installerFiles[fileIndex].shortCut.addShortcutToSystem;
      uninstallString := uninstallString + '<deletion deletionType="File" deletionPath="' +
                         IncludeTrailingPathDelimiter(getStartProgramsMenuPath) +
                         IncludeTrailingPathDelimiter(shortCutPath) +
                         installerFiles[fileIndex].shortCut.description + '.lnk'+ '"/>';
      end;
    end;
end;

function TPalmInstaller.isInstalledUser(userID: integer): boolean;
var
  ii: Integer;
begin
  result := false;
  for ii := Low(installUsers) to High(installUsers) do
    begin
    if installUsers[ii].id = userID then
      begin
      result := true;
      exit;
      end;
    end; // for
end;

function TPalmInstaller.getFileGroupCount(groupIndex: word): word;
var
  ii: Integer;
begin
  result := 0;
  for ii := low(InstallerFiles) to High(installerFiles) do
    begin
    if installerFiles[ii].fileGroup = groupIndex then
      inc(result);
    end; // for
end;

function TPalmInstaller.getSlotName(userID:LongInt; slotID: LongInt): string;
var
  size: LongInt;
  slotName: PChar;
begin
  size := MAX_PATH;
  slotName := StrAlloc(size-1);
  size := 1;
  GetSlotDisplayName(userID, slotID, slotName, size);   //first call fails, but returns correct size
  GetSlotDisplayName(userID, slotID, slotName, size);
  result := string(slotName);
  strDispose(slotName);
end;

function TPalmInstaller.fileIsSelectedForInstallation(fileIndex: integer): boolean;
var
  ii: Integer;
begin
  result := installerFiles[fileIndex].fileGroup = 0;
  try
    for ii := Low(installUsers) to High(installUsers) do
      begin
      if installUsers[ii].installFileMatrix[fileIndex] >= 1 then
        begin
        result := true;
        exit;
        end;
      end; // for
  except
    result := false;
  end;
end;

function TPalmInstaller.fileGroupIsOnlyPalmFiles(groupIndex: word): Boolean;
var
  ii: Integer;
begin
  result := false;
  for ii := low(installerFiles) to high(installerFiles) do
    begin
    if installerFiles[ii].fileGroup <> groupIndex then continue;
    if upperCase(installerFiles[ii].destination) <> '$PALM_DEVICE'  then
      begin
      exit;
      end;
    end; // for
  result := true;
end;

function TPalmInstaller.fileGroupIsCardInstallable(groupIndex: word): Boolean;
var
  ii: Integer;
begin
  result := false;
  for ii := low(installerFiles) to high(installerFiles) do
    begin
    if installerFiles[ii].fileGroup <> groupIndex then continue;
    if not installerFiles[ii].cardInstallable  then
      begin
      exit;
      end;
    end; // for
  result := true;
end;

procedure TPalmInstaller.fillCardList(userIndex: LongInt; cardList: TStringList);
var
  id:LongInt;
  count: word;
  slotList: array of LongInt;
  ii:integer;
  medType: integer;
begin
  cardList.Clear;
  getUserID(userIndex, id);
  getSlotCount(id, count);
  if count = 0 then exit;
  setLength(slotlist, count);
  if GetSlotInfo(id, slotList[0], count) <> 0 then raise
    EConduitRegistryError.Create('Unable to load slotInfo');
  for ii:= count-1 downto 0 do
    begin
    GetSlotMediaType(id, slotlist[ii], medType);
    if medType <> ExpMediaType_PlugNPlay then
      begin
      cardList.add(getSlotName(id, slotlist[ii]));
      end;
    end;
end;

function TPalmInstaller.getInstallableSlotCount(userID: integer) : word;
var
  count: word;
  slotList: array of LongInt;
  medType: integer;
  ii: integer;
begin
  result := 0;
  getSlotCount(userID, count);
  if count = 0 then exit;

  setLength(slotlist, count);
  if GetSlotInfo(userID, slotList[0], count) <> 0 then raise
    EConduitRegistryError.Create('Unable to load slotInfo');

  for ii:= count-1 downto 0 do
    begin
    GetSlotMediaType(userID, slotlist[ii], medType);
    if medType <> ExpMediaType_PlugNPlay then
      begin
      inc(result);
      end;
    end;
end;

procedure TPalmInstaller.registerConduit(settings: TConduitSettings);
var
  err: integer;
  path: string;
begin
   AddLogEntry('Beginning conduit setup API calls');

   err:=InstallCreator(PChar(settings.creatorID), CONDUIT_APPLICATION);

   if (err <> 0) and (err<>ERR_CREATORID_ALREADY_IN_USE) then
      raise EConduitRegistryError.Create('Problem Registering Conduit');

   err:=SetCreatorName(PChar(settings.creatorID), pchar(settings.fileName));
   if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Path');

   if settings.JavaSettings <> nil then
     begin
     err:=SetCreatorName(PChar(settings.creatorID), 'jsync13.dll');
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Path');
     path := settings.JavaSettings.ClassPath;
     path := StringReplace(path, '$USERINSTALLDIR', IncludeTrailingPathDelimiter(installationPath), [rfReplaceAll, rfIgnoreCase]);
     path := StringReplace(path, '$HOTSYNCDIR', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
     path := StringReplace(path, '$HOTSYNCDIR', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
     path := StringReplace(path, '$SYSDIR', IncludeTrailingPathDelimiter(getSystemPath), [rfReplaceAll, rfIgnoreCase]);
     path := StringReplace(path, '$WINDIR', IncludeTrailingPathDelimiter(getWindowsPath), [rfReplaceAll, rfIgnoreCase]);

     err := SetCreatorValueString(PChar(settings.creatorID), pchar('ClassPath13'), PChar(path));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit ClassPath');

     err := SetCreatorValueString(PChar(settings.creatorID), pchar('ClassName'), PChar(settings.JavaSettings.ClassName));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit ClassName');
     end;

   if settings.comSettings <> nil then
     begin
     err:=SetCreatorName(PChar(settings.creatorID), 'comconduit.dll');
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Path');
     err := SetCreatorValueString(PChar(settings.creatorID), pchar('COMClient'), PChar(settings.comSettings.COMClient));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit COMClient');
     end;
   err:=SetCreatorPriority(PChar(settings.creatorID), settings.cndPriority);

   if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Priority');
   if settings.cndfile <> '' then
     begin
     err:=SetCreatorFile(PChar(settings.creatorID), PChar(settings.cndFile));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit File');
     end; // if cndFile <>> ''...

   if settings.cndDirectory <> '' then
     begin
     err:=SetCreatorDirectory(PChar(settings.creatorID), PChar(settings.cndDirectory));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Directory');
     end;
   if settings.cndRemote <> '' then
     begin
     err:=SetCreatorRemote(PChar(settings.creatorID), PChar(settings.cndremote));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Remote Database');
     end;
   if settings.cndUser <> '' then
     begin
     err:=SetCreatorUser(PChar(settings.creatorID), PChar(settings.cndUser));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit User');
     end;
   if settings.cndInfo <> '' then
     begin
     err:=SetCreatorInfo(PChar(settings.creatorID), PChar(settings.cndInfo));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Info');
     end;
   if settings.cndTitle <> '' then
     begin
     err:=SetCreatorTitle(PChar(settings.creatorID), PChar(settings.cndTitle));
     if err <> 0 then raise EConduitRegistryError.Create('Problem Setting Conduit Title');
     end; // if productName <>> ''...
end;

procedure TPalmInstaller.buildUninstaller;
var
  ii: Integer;
  a: array[0..255] of char;
  regPath: TStringList;
  reg: TRegistry;
  regPathPrefix: string;
  unInstallerFileName: string;
  unInstallFile: TextFile;

begin
   if not pcFilesSelectedForInstallation then exit;

   regPath := TStringList.Create;
   AddLogEntry('loading registry for uninstall init.');
   LoadString(hInstDll, REG_PATH_ENTRY, a, sizeOf(a));
   regPathPrefix := (StrPas(a));

   for ii := Low(installerFiles) to High(installerFiles) do    // Iterate
     begin
     if installerFiles[ii].conduitSettings <> nil then
       begin
       if installerFiles[ii].conduitSettings.javaSettings <> nil then
         regPath.Add(findRegEntry(regPathPrefix, installerFiles[ii].conduitSettings.javaSettings.ClassName, ctJava))
       else if installerFiles[ii].conduitSettings.comSettings <> nil then
         regPath.Add(findRegEntry(regPathPrefix, installerFiles[ii].conduitSettings.comSettings.COMClient, ctCom))
       else
         regPath.Add(findRegEntry(regPathPrefix, installerFiles[ii].conduitSettings.fileName));
       end;
     end;    // for
   for ii:= 0 to regPath.Count-1 do
     begin
     uninstallString := uninstallString + '<deletion deletionType="RegistryKey" deletionPath="'+
                        regPath[ii] + '"/>';
     end;
   uninstallString := uninstallString + '<deletion deletionType="RegistryKey" deletionPath="'+
                        '\Software\Microsoft\Windows\CurrentVersion\Uninstall\' + productName +
                        '"/>';

   if ConduitSelectedForInstallation then
     begin
     uninstallString := uninstallString + '<hotSyncPath>' + hotSyncPath + '</hotSyncPath>';
     end;
   uninstallString := uninstallString + '</pUninstaller>';
   unInstallerFileName := IncludeTrailingPathDelimiter(getSystemPath) + productName + '.un0';
   ii := 0;
   while FileExists(uninstallerFileName) do
     begin
     inc(ii);
     unInstallerFileName := IncludeTrailingPathDelimiter(getSystemPath) + productName + '.un' +
                            trim(intToStr(ii));
     end;
   AssignFile(unInstallFile, unInstallerFileName);
   reWrite(unInstallFile);
   writeLn(unInstallFile, unInstallString);
   CloseFile(unInstallFile);
   uninstallString := IncludeTrailingPathDelimiter(getSystemPath)+extractFileName(paramStr(0)) +
                ' /u "' + unInstallerFileName + '"';
   reg := tregistry.create(key_read or key_write);
   reg.RootKey := HKEY_LOCAL_MACHINE;
   AddLogEntry('Applying uninstall string to registry');
   reg.OpenKey('\Software\Microsoft\Windows\CurrentVersion\Uninstall\' + productName, true);
   reg.WriteString('UninstallString', uninstallString);
   reg.WriteString('DisplayName', productName);
   reg.Free;
   regPath.Free;

   AddLogEntry('copying installer to system directory');
   if fileVersionIsHigher(paramStr(0), IncludeTrailingPathDelimiter(getSystemPath)+extractFileName(paramStr(0))) then
     begin
     FileCopy(paramStr(0), IncludeTrailingPathDelimiter(getSystemPath)+extractFileName(paramStr(0)));
     end;
   increaseUsageCount(IncludeTrailingPathDelimiter(getSystemPath)+extractFileName(paramStr(0)));

end;

procedure TPalmInstaller.applyCustomRegistryChanges;
var
  ii: Integer;
  regContents: TStringList;
  regPath: string;
  a: array[0..255] of char;
  regFile: string;
  reg: TREgistry;
begin
  AddLogEntry('Applying custom registry changes');
  if registryFiles.Count <> registryConduits.Count then
    begin
    raise Exception.Create('Registry Configuration Error');
    exit;
    end;
  for ii := 0 to registryFiles.Count - 1 do
    begin
    regFile := registryFiles[ii];
    if (regFile = '') or (not fileExists(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0)))+regFile))then
      continue;
    if registryConduits[ii] <> '' then
      begin
      //apply any custom registry entries
      //(replaces <INSTALL_KEY> with the installed Key)
      LoadString(hInstDll, REG_PATH_ENTRY, a, sizeOf(a));
      regPath := (StrPas(a));
      regPath := findRegEntry(regPath, registryConduits[ii]);
      AddLogEntry('Adding custom registry entries');
      regContents := TStringList.Create;

      regContents.LoadFromFile(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0)))+regFile);
      regContents.Text := StringReplace(regContents.Text, '<INSTALL_KEY>', 'HKEY_CURRENT_USER' + regPath, [rfReplaceAll, rfIgnoreCase]);
      regContents.SaveToFile(IncludeTrailingPathDelimiter(getTempDir) + 'tmpRegEntry.reg');
      regContents.Free;
      end
    else
      begin
      FileCopy(IncludeTrailingPathDelimiter(extractFilePath(paramStr(0)))+regFile, IncludeTrailingPathDelimiter(getTempDir) + 'tmpRegEntry.reg');
      end;

    AddLogEntry('launching regedit');
    shellexec(nil, PChar('regedit.exe'), PChar('/s ' + IncludeTrailingPathDelimiter(getTempDir) + 'tmpRegEntry.reg'), nil, SW_SHOWDEFAULT, true);
    AddLogEntry('deleting temp reg file');
    killFile('tmpRegEntry.reg');
    end; // for
  reg := TRegistry.Create(key_read or key_write);
  for ii := Low(regChanges) to High(regChanges) do
    begin
    reg.RootKey := regChanges[ii].rootKey;
    if regChanges[ii].hivePath[1] = '\' then
      reg.OpenKey(regChanges[ii].hivePath, true)
    else
      reg.OpenKey('\' + regChanges[ii].hivePath, true);
    if regChanges[ii].valueType = rtString then
      begin
      reg.WriteString(regChanges[ii].name, regChanges[ii].stringValue);
      end
    else
      begin
      reg.WriteInteger(regChanges[ii].name, regChanges[ii].intValue);
      end;
    end; // for
  reg.Free;
end;

procedure TPalmInstaller.runPostInstallExes;
var
  ii, jj: Integer;
  exeParams: string;
  users: string;
begin
  for ii := Low(installerFiles) to High(installerFiles) do
    begin
    if not installerFiles[ii].executeAfterInstall then
      continue;
    AddLogEntry('executing post install executable ' + installerFiles[ii].fileName);
    exeParams := installerFiles[ii].executionParameters;
    for jj := low(installUsers) to high(installUsers) do
      begin
      users := users + installUsers[jj].displayName + ',';
      end;
    if users <> '' then
      users := copy(users, 0, length(users) -1);
    exeParams := StringReplace(exeParams, '{?username?}', users, [rfReplaceAll, rfIgnoreCase]);
    exeParams := StringReplace(exeParams, '{?conduitPath?}', hotSyncPath, [rfReplaceAll, rfIgnoreCase]);
    if installerFiles[ii].doNotInstall then
      shellexec(nil, PChar(installerFiles[ii].fileName), PChar(exeParams), nil, SW_SHOWDEFAULT, true)
    else
      shellexec(nil, PChar(IncludeTrailingPathDelimiter(installationPath) + installerFiles[ii].fileName),
                     PChar(exeParams), nil, SW_SHOWDEFAULT, true);
    end; // for
end;

procedure TPalmInstaller.restartHotSync;
var
  hsLibName: string;
  HsSetAppStatus: FUNCTION(statusType: longInt; flags:longInt):LongInt stdcall;
  hHSDll : HModule;
begin
  hsLibName:= IncludeTrailingPathDelimiter(hotSyncPath) + 'hsAPI.dll';
  hHSDll := LoadLibraryEx(PChar(hsLibName),0,0);
  AddLogEntry('restarting hotsync');
  HsSetAppStatus:= GetProcAddress(hHSDll, 'HsSetAppStatus');
  HsSetAppStatus(2, 0);
  AddLogEntry('HotSync Restarted');
end;

function TPalmInstaller.execute: boolean;
var
  frmInstallationWizard: TfrmInstallationWizard;
  path: string;
begin
  AddLogEntry('executing installation');
  result := false;
  path := installationPath;
  path := StringReplace(path, '$HOTSYNCDIR\', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$SYSDIR\', IncludeTrailingPathDelimiter(getSystemPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$WINDIR\', IncludeTrailingPathDelimiter(getWindowsPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$PROGFILESDIR\', IncludeTrailingPathDelimiter(getProgramFilesPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$HOTSYNCDIR', IncludeTrailingPathDelimiter(hotSyncPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$SYSDIR', IncludeTrailingPathDelimiter(getSystemPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$WINDIR', IncludeTrailingPathDelimiter(getWindowsPath), [rfReplaceAll, rfIgnoreCase]);
  path := StringReplace(path, '$PROGFILESDIR', IncludeTrailingPathDelimiter(getProgramFilesPath), [rfReplaceAll, rfIgnoreCase]);
  installationPath := path;

  AddLogEntry('Creating form');
  frmInstallationWizard := TfrmInstallationWizard.Create(Application);
  frmInstallationWizard.installerSettings := self;
  frmInstallationWizard.lblIntroTitle.Caption := frmInstallationWizard.lblIntroTitle.Caption + productName;
  frmInstallationWizard.lblIntroDetails.Caption := frmInstallationWizard.lblIntroDetails.Caption + productName;
  frmInstallationWizard.pnlIntroContainer.BringToFront;
  if (not runSilent) or preInstallationMessageOverrideSilentMode or
//               postInstallationMessageOverrideSilentMode or
               userPromptOverrideSilentMode then
    begin
    frmInstallationWizard.ShowModal;
    end
  else
    begin
    AddLogEntry('Silent mode with overrides');
    frmInstallationWizard.currentPage := ipProgress;
    frmInstallationWizard.jumpToProgress;
    frmInstallationWizard.ShowModal;
    end;
  frmInstallationWizard.Free;
  if forceReboot then
    begin
    Application.MessageBox('Installation successful.  Your computer needs ' +
                           'to be rebooted to complete the installation. Press OK to proceed',
                           'Reboot Needed', MB_OK);
    NTSetPrivilege('SeShutdownPrivilege', true);
    ExitWindowsEx(EWX_REBOOT + EWX_FORCE,0);
    end
  else if FRebootNeeded then
    begin
    if Application.MessageBox('Install successful.  Your computer needs ' +
                           'to be rebooted to complete the installation.  Do you want to restart now?',
                           'Reboot Needed', MB_YESNO or MB_ICONQUESTION) = idYes
                           then
      begin
      NTSetPrivilege('SeShutdownPrivilege', true);
      ExitWindowsEx(EWX_REBOOT + EWX_FORCE,0);
      end;
    end;

end;

function TPalmInstaller.getHotSyncUserCount: integer;
begin
  result := getUserCount;
end;

end.
